"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007
    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from collections import defaultdict
from dataclasses import dataclass

import geopandas as gpd
import networkx as nx
from geopandas import GeoDataFrame
from numpy import int64 as npInt64
from shapely.geometry import LineString, Point
from snkit.network import Network as SnkitNetwork

NxGraph = nx.Graph | nx.MultiGraph | nx.MultiDiGraph


@dataclass(kw_only=True)
class NxToSnkitNetworkConverter:
    """
    Class responsible to convert a `networkx.MultiGraph` into
    a matching `snkit.network.Network`.
    """

    networkx_graph: NxGraph
    node_id_column_name: str = "id"
    edge_from_id_column_name: str = "from_id"
    edge_to_id_column_name: str = "to_id"

    def convert(self) -> SnkitNetwork:
        """
        Converts a regular `NetworkX.graph` into a `snkit.network.Network` object.

        Returns:
            SnkitNetwork: The resulting `snkit.network.Network` converted object.
        """
        # Extract graph values
        _crs = self.networkx_graph.graph.get("crs", None)

        # Create new network
        snkit_network = SnkitNetwork()
        node_attributes = [
            {self.node_id_column_name: node, **data}
            for node, data in self.networkx_graph.nodes(data=True)
        ]
        snkit_network.nodes = GeoDataFrame(node_attributes)
        snkit_network.nodes = self._check_and_create_node_geometries(
            snkit_network.nodes
        )
        snkit_network.nodes.set_geometry("geometry", inplace=True, crs=_crs)

        edge_attributes = [
            {self.edge_from_id_column_name: u, self.edge_to_id_column_name: v, **data}
            for u, v, data in self.networkx_graph.edges(data=True)
        ]
        snkit_network.edges = GeoDataFrame(edge_attributes)
        snkit_network = self._check_and_create_edge_geometries(snkit_network)
        snkit_network.edges.set_geometry("geometry", inplace=True, crs=_crs)

        # Set network CRS to default_crs
        snkit_network.set_crs(_crs)

        # Checks
        snkit_network = self._check_edge_ids(snkit_network)
        snkit_network = self._get_nodes_degree(snkit_network)

        # Return converted and validated network
        return snkit_network

    def _check_edge_ids(self, network: SnkitNetwork) -> SnkitNetwork:
        if not id in network.edges.columns:
            network.edges["id"] = network.edges.index
        return network

    def _get_nodes_degree(self, network: SnkitNetwork) -> SnkitNetwork:
        def _calculate_degree(snkit_network: SnkitNetwork) -> dict:
            degrees = defaultdict(int)

            from_ids = snkit_network.edges["from_id"].to_numpy(dtype=npInt64)
            for from_id in from_ids:
                degrees[from_id] += 1

            to_ids = snkit_network.edges["to_id"].to_numpy(dtype=npInt64)
            for to_id in to_ids:
                degrees[to_id] += 1

            return degrees

        degrees = _calculate_degree(network)
        network.nodes["degree"] = network.nodes["id"].apply(
            lambda node_id: degrees.get(node_id, 0)
        )
        return network

    def _check_and_create_node_geometries(
        self,
        geo_dataframe: gpd.GeoDataFrame,
    ) -> gpd.GeoDataFrame:
        """
        Check if there is a geometry column in network.nodes.
        If not, check if both 'x' and 'y' columns are present.
        If both are present, create Point geometries for each row.
        If either 'x' or 'y' is missing, raise a ValueError.

        Parameters:
        gdf (snkit.network.Network): The gdf object containing nodes as a GeoDataFrame.

        Raises:
        ValueError: If neither geometry column nor both 'x' and 'y' columns are present in network.nodes.
        """
        if "geometry" in geo_dataframe.columns:
            return geo_dataframe

        if "x" in geo_dataframe.columns and "y" in geo_dataframe.columns:
            geo_dataframe["geometry"] = geo_dataframe.apply(
                lambda row: Point(row["x"], row["y"]), axis=1
            )
            return geo_dataframe
        else:
            raise ValueError(
                "The network nodes must contain either a 'geometry' column or both 'x' and 'y' columns."
            )

    def _check_and_create_edge_geometries(
        self,
        network: SnkitNetwork,
    ) -> SnkitNetwork:
        """
        Creates a GEOMETRY attribute for each edge in the graph using the geometries of the nodes.

        Parameters:
        G (nx.Graph): The NetworkX graph with nodes having geometries.

        Returns:
        nx.Graph: The NetworkX graph with edges having GEOMETRY attributes.
        """

        def create_linestring(row):
            from_geom = node_geometries.get(row["from_id"])
            to_geom = node_geometries.get(row["to_id"])
            if from_geom is None or to_geom is None:
                raise ValueError(
                    f"Geometry missing for from_id {row['from_id']} or to_id {row['to_id']}."
                )
            return LineString([from_geom, to_geom])

        if "geometry" in network.edges.columns:
            return network
        # Check if nodes have geometries
        if not ("geometry" in network.nodes.columns):
            network.nodes = self._check_and_create_node_geometries(network.nodes)

        # Convert nodes to a dictionary for fast lookup
        node_geometries = network.nodes.set_index("id")["geometry"].to_dict()

        # Apply the function to create the geometry column
        network.edges["geometry"] = network.edges.apply(create_linestring, axis=1)

        return network



import copy
from pathlib import Path

import networkx as nx
import numpy as np
import pandas as pd
from geopandas import GeoDataFrame, read_feather, read_file

from ra2ce.analysis.analysis_config_data.analysis_config_data import (
    AnalysisSectionLosses,
)
from ra2ce.analysis.analysis_input_wrapper import AnalysisInputWrapper
from ra2ce.analysis.losses.analysis_losses_protocol import AnalysisLossesProtocol
from ra2ce.analysis.losses.optimal_route_origin_destination import (
    OptimalRouteOriginDestination,
)
from ra2ce.network.graph_files.graph_file import GraphFile
from ra2ce.network.hazard.hazard_names import HazardNames
from ra2ce.network.network_config_data.network_config_data import (
    OriginsDestinationsSection,
)


class MultiLinkOriginDestination(AnalysisLossesProtocol):
    analysis: AnalysisSectionLosses
    graph_file_hazard: GraphFile
    input_path: Path
    static_path: Path
    output_path: Path
    hazard_names: HazardNames
    origins_destinations: OriginsDestinationsSection
    _analysis_input: AnalysisInputWrapper

    def __init__(
        self,
        analysis_input: AnalysisInputWrapper,
    ) -> None:
        self.analysis = analysis_input.analysis
        self.graph_file_hazard = analysis_input.graph_file_hazard
        self.input_path = analysis_input.input_path
        self.static_path = analysis_input.static_path
        self.output_path = analysis_input.output_path
        self.hazard_names = analysis_input.hazard_names
        self.origins_destinations = analysis_input.origins_destinations
        self._analysis_input = analysis_input

    @staticmethod
    def extract_od_nodes_from_graph(
        graph: nx.MultiGraph,
    ) -> list[tuple[str, str]]:
        """
        Extracts all Origin - Destination nodes from the graph, prevents from entries
        with list of nodes for a node.

        Args:
            graph (nx.MultiGraph): Graph containing origin-destination nodes.

        Returns:
            list[tuple[str, str]]]: List containing tuples of origin - destination node combinations.
        """
        _od_nodes = []
        for n, v in graph.nodes(data=True):
            if "od_id" not in v:
                continue
            _o_node_list = list(map(lambda x: (n, x), v["od_id"].split(",")))
            _od_nodes.extend(_o_node_list)
        return _od_nodes

    def _get_origin_destination_pairs(
        self, graph: nx.MultiGraph
    ) -> list[tuple[str, str], tuple[str, str]]:
        od_path = self.static_path.joinpath(
            "output_graph", "origin_destination_table.feather"
        )
        od = read_feather(od_path)
        od_pairs = [
            (a, b)
            for a in od.loc[od["o_id"].notnull(), "o_id"]
            for b in od.loc[od["d_id"].notnull(), "d_id"]
        ]
        all_nodes = self.extract_od_nodes_from_graph(graph)
        od_nodes = []
        for aa, bb in od_pairs:
            # it is possible that there are multiple origins/destinations at the same 'entry-point' in the road
            od_nodes.append(
                (
                    [
                        (n, n_name)
                        for n, n_name in all_nodes
                        if (n_name == aa) | (aa in n_name)
                    ][0],
                    [
                        (n, n_name)
                        for n, n_name in all_nodes
                        if (n_name == bb) | (bb in n_name)
                    ][0],
                )
            )
        return od_nodes

    def multi_link_origin_destination(
        self, graph: nx.MultiGraph, analysis: AnalysisSectionLosses
    ) -> GeoDataFrame:
        """Calculates the connectivity between origins and destinations"""
        od_nodes = self._get_origin_destination_pairs(graph)

        all_results = []
        for hazard in self.hazard_names.names:
            hazard_name = self.hazard_names.get_name(hazard)

            graph_hz = copy.deepcopy(graph)

            # Check if the o/d pairs are still connected while some links are disrupted by the hazard(s)
            edges_remove = [
                e for e in graph.edges.data(keys=True) if hazard_name in e[-1]
            ]
            edges_remove = [e for e in edges_remove if (e[-1][hazard_name] is not None)]
            edges_remove = [
                e
                for e in edges_remove
                if (e[-1][hazard_name] > float(analysis.threshold))
                & ("bridge" not in e[-1])
            ]
            graph_hz.remove_edges_from(edges_remove)

            # convert the networkx graph to igraph object to speed up the route finding algorithm
            # igraph_hz = ig.Graph.from_networkx(igraph_hz)

            # Find the routes
            od_routes = OptimalRouteOriginDestination.find_route_ods(
                graph_hz, od_nodes, analysis.weighing
            )
            od_routes["hazard"] = hazard_name
            all_results.append(od_routes)

        return pd.concat(all_results, ignore_index=True)

    def multi_link_origin_destination_impact(
        self, gdf: GeoDataFrame, gdf_ori: GeoDataFrame
    ) -> tuple[pd.DataFrame, GeoDataFrame]:
        """Calculates some default indicators that quantify the impacts of disruptions to origin-destination flows
        The function outputs the following file:

        1. gdf_ori (multi_link_origin_destination_impact.csv), containing the following information:
            - origin and destination node
            - length: initial shortest path length before disruptions
            - length_hazardName: shortest path length after disruption
            - diff_length_hazardName: increase in shortest path length after disruption (length_hazardName - length)
            - diff_length_hazardName_pc: same as above, but as a fraction of initial length

        2. diff_df (multi_link_origin_destination_impact_summary.csv), containing the following information:
            - hazard: hazard name
            - od_disconnected_abs: number of OD disconnected
            - od_disconnected_pc (%): percentage of OD disconnected
            - origin_disconnected_abs: number of origin points disconnected
            - origin_disconnected_pc (%): percentage of origin points disconnected
            - destination_disconnected_abs: number of destination points disconnected
            - destination_disconnected_pc (%): percentage of destination points disconnected
            - max_increase_abs: maximum increase in travel length across all OD pairs
            - mean_increase_abs: mean increase in travel length across all OD pairs
            - median_increase_abs: median increase in travel length across all OD pairs
            - max_increase_pc, mean_increase_pc, median_increase_pc (%): same as above three, but as a percentage relative to no-hazard
        """

        hazard_list = np.unique(gdf["hazard"])

        # calculate number of disconnected origin, destination, and origin-destination pair
        # TODO: there seems to be an issue in calculating origin_count and destination_count where origin and destination nodes are the same, e.g., A_25,B_1
        gdf["OD"] = gdf["origin"] + gdf["destination"]
        gdf_ori["OD"] = gdf_ori["origin"] + gdf_ori["destination"]
        # origin
        gdf_ori["origin_count"] = gdf_ori["origin"].apply(lambda x: len(x.split(",")))
        init_origins = gdf_ori.groupby("origin")["origin_count"].sum()
        del gdf_ori["origin_count"]
        # destination
        gdf_ori["destination_count"] = gdf_ori["destination"].apply(
            lambda x: len(x.split(","))
        )
        init_destinations = gdf_ori.groupby("destination")["destination_count"].sum()
        del gdf_ori["destination_count"]
        # od pairs
        init_od_pairs = init_origins * init_destinations
        abs_od_disconnected = []
        share_od_disconnected = []
        abs_origin_disconnected = []
        share_origin_disconnected = []
        abs_destination_disconnected = []
        share_destination_disconnected = []
        for hz in hazard_list:
            gdf_ = gdf.loc[gdf["hazard"] == hz]

            gdf_["origin_count"] = gdf_["origin"].apply(lambda x: len(x.split(",")))
            remaining_origins = gdf_.groupby("origin")["origin_count"].sum()
            del gdf_["origin_count"]
            diff_origins = init_origins - remaining_origins
            abs_origin_disconnected.append(diff_origins)
            share_origin_disconnected.append(100 * diff_origins / init_origins)

            gdf_["destination_count"] = gdf_["destination"].apply(
                lambda x: len(x.split(","))
            )
            remaining_destinations = gdf_.groupby("destination")[
                "destination_count"
            ].sum()
            del gdf_["destination_count"]
            diff_destinations = init_destinations - remaining_destinations
            abs_destination_disconnected.append(diff_destinations)
            share_destination_disconnected.append(
                100 * diff_destinations / init_destinations
            )

            remaining_od_pairs = remaining_origins * remaining_destinations
            diff_od_pairs = init_od_pairs - remaining_od_pairs
            abs_od_disconnected.append(diff_od_pairs)
            share_od_disconnected.append(100 * diff_od_pairs / init_od_pairs)

        # calculate change in travel time/distance
        max_increase_abs = []
        min_increase_abs = []
        mean_increase_abs = []
        median_increase_abs = []
        max_increase_pc = []
        min_increase_pc = []
        mean_increase_pc = []
        median_increase_pc = []
        for hz in hazard_list:
            gdf_ = gdf.loc[gdf["hazard"] == hz][["OD", "length"]]
            gdf_.columns = ["OD", "length_" + hz]
            gdf_ori = gdf_ori.merge(gdf_, how="left", on="OD")
            gdf_ori.drop_duplicates(subset="OD", inplace=True)
            gdf_ori["diff_length_" + hz] = gdf_ori["length_" + hz] - gdf_ori["length"]
            gdf_ori["diff_length_" + hz + "_pc"] = (
                100 * gdf_ori["diff_length_" + hz] / gdf_ori["length"]
            )

            max_increase_abs.append(np.nanmax(gdf_ori["diff_length_" + hz]))
            min_increase_abs.append(np.nanmin(gdf_ori["diff_length_" + hz]))
            mean_increase_abs.append(np.nanmean(gdf_ori["diff_length_" + hz]))
            median_increase_abs.append(np.nanmedian(gdf_ori["diff_length_" + hz]))

            max_increase_pc.append(np.nanmax(gdf_ori["diff_length_" + hz + "_pc"]))
            min_increase_pc.append(np.nanmin(gdf_ori["diff_length_" + hz + "_pc"]))
            mean_increase_pc.append(np.nanmean(gdf_ori["diff_length_" + hz + "_pc"]))
            median_increase_pc.append(
                np.nanmedian(gdf_ori["diff_length_" + hz + "_pc"])
            )

        diff_df = pd.DataFrame()
        diff_df["hazard"] = hazard_list

        diff_df["od_disconnected_abs"] = abs_od_disconnected
        diff_df["od_disconnected_pc (%)"] = share_od_disconnected
        diff_df["origin_disconnected_abs"] = abs_origin_disconnected
        diff_df["origin_disconnected_pc (%)"] = share_origin_disconnected
        diff_df["destination_disconnected_abs"] = abs_destination_disconnected
        diff_df["destination_disconnected_pc (%)"] = share_destination_disconnected

        diff_df["max_increase_abs"] = max_increase_abs
        diff_df["min_increase_abs"] = min_increase_abs
        diff_df["mean_increase_abs"] = mean_increase_abs
        diff_df["median_increase_abs"] = median_increase_abs
        diff_df["max_increase_pc (%)"] = max_increase_pc
        diff_df["min_increase_pc (%)"] = min_increase_pc
        diff_df["mean_increase_pc (%)"] = mean_increase_pc
        diff_df["median_increase_pc (%)"] = median_increase_pc

        return diff_df, gdf_ori

    def multi_link_origin_destination_regional_impact(
        self, gdf_ori: GeoDataFrame
    ) -> tuple[GeoDataFrame, GeoDataFrame]:
        """
        Aggregation of the impacts of disruptions at region level
        Users need to specify 'region' and 'region_var' attributes in the network.ini file
        See the Pontianak case study for an example

        The function outputs the following files:
        1. multi_link_origin_destination_regional_impact.csv
            Impacts of disruption aggregated for each origin node. Region information (to which region an origin node belongs) is retained

        2. multi_link_origin_destination_regional_impact_summary.csv
            Impacts of disruption aggregated for each region.

        In both files, the following information is stored:
            - init_length: initial average length to all destination nodes (for each origin node and for each region)
            - init_destination: initial number of destination nodes
            - hazardName_pc_increase: average increase in travel time to all destination nodes (percentage relative to initial travel time, for each origin node and for each region)
            - hazardName_pc_disconnect: average number of OD pairs disconnected (relative to initial number of OD pairs, for each origin node and for each region)
        """

        gdf_ori_ = gdf_ori.copy()

        # read origin points
        origin_fn = Path(self.static_path).joinpath(
            "output_graph", "origin_destination_table.gpkg"
        )
        origin = read_file(origin_fn, engine="pyogrio")
        index = [isinstance(x, str) for x in origin["o_id"]]
        origin = origin[index]
        origin.reset_index(inplace=True, drop=True)

        # record where each origin point resides
        origin_mapping = {}
        for o in np.unique(origin["o_id"]):
            r = origin.loc[origin["o_id"] == o, "region"].values[0]
            origin_mapping.update({o: r})

        # record impact to each region
        origin_impact_master = pd.DataFrame()
        for r in np.unique(origin["region"]):
            origin_points = list(origin.loc[origin["region"] == r, "o_id"].values)
            for o in origin_points:
                origin_impact_tosave = pd.DataFrame()
                origin_impact_tosave.loc[0, "o_id"] = o
                origin_impact_tosave.loc[0, "region"] = r

                origin_impact = gdf_ori_.loc[gdf_ori_["origin"].str.contains(o)]

                # initial condition
                origin_impact_tosave.loc[0, "init_length"] = np.mean(
                    origin_impact["length"]
                )
                origin_impact_tosave.loc[0, "init_destination"] = len(
                    np.unique(origin_impact["destination"])
                )

                # impact of each hazard
                for col in origin_impact.columns:
                    if "_pc" in col:
                        delta = np.nanmean(origin_impact[col])
                        if delta < 0:
                            delta = 0
                        origin_impact_tosave.loc[0, col[12:] + "_increase"] = delta

                        disconnected = origin_impact[col].isna().sum()
                        origin_impact_tosave.loc[0, col[12:] + "_disconnect"] = (
                            100
                            * disconnected
                            / origin_impact_tosave["init_destination"].values[0]
                        )

                origin_impact_master = origin_impact_master.append(origin_impact_tosave)

        region_impact_master = origin_impact_master[origin_impact_master.columns[1:]]
        region_impact_master = region_impact_master.groupby(by="region").mean()

        return origin_impact_master, region_impact_master

    def execute(self) -> GeoDataFrame:
        _output_path = self.output_path.joinpath(self.analysis.analysis.config_value)
        gdf = self.multi_link_origin_destination(
            self.graph_file_hazard.get_graph(), self.analysis
        )
        self._analysis_input.graph_file = self._analysis_input.graph_file_hazard
        gdf_not_disrupted = OptimalRouteOriginDestination(
            self._analysis_input
        ).execute()
        (
            disruption_impact_df,
            gdf_ori,
        ) = self.multi_link_origin_destination_impact(gdf, gdf_not_disrupted)
        try:
            assert self.origins_destinations.region
            (
                regional_impact_df,
                regional_impact_summary_df,
            ) = self.multi_link_origin_destination_regional_impact(gdf_ori)
            impact_csv_path = _output_path.joinpath(
                (self.analysis.name.replace(" ", "_") + "_regional_impact.csv"),
            )
            regional_impact_df.to_csv(impact_csv_path, index=False)
            impact_csv_path = _output_path.joinpath(
                (self.analysis.name.replace(" ", "_") + "_regional_impact_summary.csv"),
            )
            regional_impact_summary_df.to_csv(impact_csv_path)
        except Exception:
            pass
        impact_csv_path = _output_path.joinpath(
            (self.analysis.name.replace(" ", "_") + "_impact.csv"),
        )
        del gdf_ori["geometry"]
        gdf_ori.to_csv(impact_csv_path, index=False)
        impact_csv_path = _output_path.joinpath(
            (self.analysis.name.replace(" ", "_") + "_impact_summary.csv"),
        )
        disruption_impact_df.to_csv(impact_csv_path, index=False)

        return gdf



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007

    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
from __future__ import annotations

from geopandas import GeoDataFrame

from ra2ce.analysis.analysis_config_data.enums.damage_curve_enum import DamageCurveEnum
from ra2ce.analysis.damages.damage_calculation.damage_network_base import (
    DamageNetworkBase,
)


class DamageNetworkEvents(DamageNetworkBase):
    """A road network gdf with EVENT-BASED hazard data stored in it, and for which damages can be calculated

    @Author: Kees van Ginkel

    Mandatory attributes:
        *self.events* (set)  : all available unique events
        *self.stats* (set)   : the available statistics
    """

    def __init__(
        self,
        road_gdf: GeoDataFrame,
        val_cols: list[str],
        representative_damage_percentage: float,
    ):
        # Construct using the parent class __init__
        super().__init__(road_gdf, val_cols, representative_damage_percentage)
        self.events = set([x.split("_")[1] for x in val_cols])  # set of unique events

        if not any(self.events):
            raise ValueError("No event cols present in hazard data")

    ### Controler for Event-based damage calculation
    def main(self, damage_function: DamageCurveEnum, manual_damage_functions=None):
        assert len(self.events) > 0, "no return periods identified"
        assert "me" in self.stats, "mean water depth (key: me) is missing"
        assert "fr" in self.stats, "inundated fraction (key: fr) is missing"

        self.do_cleanup_and_mask_creation()

        if damage_function == DamageCurveEnum.HZ:
            self.calculate_damage_HZ(events=self.events)

        if damage_function == DamageCurveEnum.OSD:
            self.calculate_damage_OSdaMage(events=self.events)

        if damage_function == DamageCurveEnum.MAN:
            self.calculate_damage_manual_functions(
                events=self.events, manual_damage_functions=manual_damage_functions
            )


# class DamageNetworkEventsBuilder:
#     @staticmethod
#     def from_csv_file(file_path) -> DamageNetworkEvents:
#         pass
#
#     @staticmethod
#     def from_pandas(pd_dataframe) -> DamageNetworkEvents:
#         pass



from typing import Any, Protocol, runtime_checkable


@runtime_checkable
class WeighingAnalysisProtocol(Protocol):
    edge_data: dict[str, Any]

    def get_current_value(self) -> float:
        """
        Gets the current distance/time of the edge.
        If the edge has no distance/time attribute, it is calculated and added to the edge.

        Returns:
            float: Current distance/time value.
        """



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007

    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import logging

import geopandas as gpd
from geopandas import GeoDataFrame
from networkx import MultiGraph

from ra2ce.common.io.readers import GraphPickleReader
from ra2ce.network.network_config_data.enums.source_enum import SourceEnum
from ra2ce.network.network_config_data.network_config_data import NetworkConfigData
from ra2ce.network.network_wrappers.network_wrapper_protocol import (
    NetworkWrapperProtocol,
)
from ra2ce.network.network_wrappers.osm_network_wrapper.osm_network_wrapper import (
    OsmNetworkWrapper,
)
from ra2ce.network.network_wrappers.shp_network_wrapper import ShpNetworkWrapper
from ra2ce.network.network_wrappers.trails_network_wrapper import TrailsNetworkWrapper
from ra2ce.network.network_wrappers.vector_network_wrapper import VectorNetworkWrapper


class NetworkWrapperFactory(NetworkWrapperProtocol):
    def __init__(self, config_data: NetworkConfigData) -> None:
        self._config_data = config_data

    def _any_cleanup_enabled(self) -> bool:
        _cleanup = self._config_data.cleanup
        return (
            _cleanup.snapping_threshold
            or _cleanup.pruning_threshold
            or _cleanup.merge_lines
            or _cleanup.cut_at_intersections
        )

    def get_network(self) -> tuple[MultiGraph, GeoDataFrame]:
        logging.info("Start creating a network from the submitted shapefile.")
        source = self._config_data.network.source
        if source == SourceEnum.SHAPEFILE:
            if self._any_cleanup_enabled():
                return ShpNetworkWrapper(self._config_data).get_network()
            return VectorNetworkWrapper(self._config_data).get_network()
        elif source == SourceEnum.OSB_BPF:
            return TrailsNetworkWrapper(self._config_data).get_network()
        elif source == SourceEnum.OSM_DOWNLOAD:
            return OsmNetworkWrapper(self._config_data).get_network()
        elif source == SourceEnum.PICKLE:
            logging.info("Start importing a network from pickle")
            base_graph = GraphPickleReader().read(
                self._config_data.output_graph_dir.joinpath("base_graph.p")
            )
            network_gdf = gpd.read_feather(
                self._config_data.output_graph_dir.joinpath("base_network.feather")
            )
            return base_graph, network_gdf
        raise ValueError(f"Source {source} is not supported.")



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007

    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import logging
from typing import Any


class ValidationReport:
    _errors: list[str]
    _warns: list[str]

    def __init__(self) -> None:
        self._errors = []
        self._warns = []

    def error(self, error_mssg: str) -> None:
        logging.error(error_mssg)
        self._errors.append(error_mssg)

    def warn(self, warn_mssg: str) -> None:
        logging.warning(warn_mssg)
        self._warns.append(warn_mssg)

    def is_valid(self) -> bool:
        return not any(self._errors)

    def merge(self, with_report: Any) -> None:
        """
        Merges a given report with the current one.

        Args:
            with_report (Any): ValidationReport that will be merged.
        """
        self._errors.extend(with_report._errors)
        self._warns.extend(with_report._warns)



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007

    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from typing import Any

from ra2ce.analysis.analysis_config_data.analysis_config_data import (
    AnalysisConfigData,
    DamagesAnalysisNameList,
    LossesAnalysisNameList,
)
from ra2ce.common.validation.ra2ce_validator_protocol import Ra2ceIoValidator
from ra2ce.common.validation.validation_report import ValidationReport
from ra2ce.configuration.ra2ce_enum_base import Ra2ceEnumBase
from ra2ce.network.network_config_data.network_config_data_validator import (
    NetworkDictValues,
)

AnalysisNetworkDictValues = NetworkDictValues | {
    "analysis": LossesAnalysisNameList + DamagesAnalysisNameList
}


class AnalysisConfigDataValidator(Ra2ceIoValidator):
    def __init__(self, config_data: AnalysisConfigData) -> None:
        self._config = config_data

    def _validate_header(self, header: Any) -> ValidationReport:
        _report = ValidationReport()

        if isinstance(header, list):
            for _item in header:
                _report.merge(self._validate_header(_item))
        else:
            # check keys with predescribed values
            for key, value in header.__dict__.items():
                if not value:
                    continue
                if isinstance(value, Ra2ceEnumBase):
                    # enumerations
                    _expected_values_list = value.list_valid_options()
                else:
                    # other items with limited value options (should become enumerations)
                    if key not in AnalysisNetworkDictValues.keys():
                        continue
                    _expected_values_list = AnalysisNetworkDictValues[key]
                if value not in _expected_values_list:
                    _report.error(
                        f"Wrong input to property [ {key} ]; has to be one of: {_expected_values_list}."
                    )

        return _report

    def _validate_headers(self, required_headers: list[str]) -> ValidationReport:
        _report = ValidationReport()
        _available_keys = self._config.__dict__.keys()

        def _check_header(header: str) -> None:
            """
            Check if the section is provided and non-empty

            Args:
                header (str): section name
            """
            if header not in _available_keys or not getattr(self._config, header):
                _report.error(
                    f"Property [ {header} ] is not configured. Add property [ {header} ] to the *.ini file. "
                )

        list(map(_check_header, required_headers))
        if not _report.is_valid():
            return _report

        # check if properties have correct input
        # TODO: Decide whether also the non-used properties must be checked or those are not checked

        for header in required_headers:
            # Now check the parameters per configured item.
            _attr = getattr(self._config, header)
            if not _attr:
                continue
            else:
                _report.merge(self._validate_header(_attr))

        if not _report.is_valid():
            _report.error(
                "There are inconsistencies in the *.ini file. Please consult the log file for more information: {}.".format(
                    self._config.output_path.joinpath("RA2CE.log")
                )
            )

        return _report

    def validate(self) -> ValidationReport:
        _report = ValidationReport()
        _required_headers = ["project", "analyses"]

        _report.merge(self._validate_headers(_required_headers))
        return _report



from pathlib import Path

import networkx as nx
import pandas as pd
from geopandas import GeoDataFrame, read_feather
from shapely.geometry import LineString, MultiLineString
from tqdm import tqdm

from ra2ce.analysis.analysis_config_data.analysis_config_data import (
    AnalysisSectionLosses,
)
from ra2ce.analysis.analysis_config_data.enums.weighing_enum import WeighingEnum
from ra2ce.analysis.analysis_input_wrapper import AnalysisInputWrapper
from ra2ce.analysis.losses.analysis_losses_protocol import AnalysisLossesProtocol
from ra2ce.analysis.losses.traffic_analysis.traffic_analysis_factory import (
    TrafficAnalysisFactory,
)
from ra2ce.network.graph_files.graph_file import GraphFile
from ra2ce.network.hazard.hazard_names import HazardNames
from ra2ce.network.network_config_data.network_config_data import (
    OriginsDestinationsSection,
)


class OptimalRouteOriginDestination(AnalysisLossesProtocol):
    analysis: AnalysisSectionLosses
    graph_file: GraphFile
    input_path: Path
    static_path: Path
    output_path: Path
    hazard_names: HazardNames
    origins_destinations: OriginsDestinationsSection

    def __init__(
        self,
        analysis_input: AnalysisInputWrapper,
    ) -> None:
        self.analysis = analysis_input.analysis
        self.graph_file = analysis_input.graph_file
        self.input_path = analysis_input.input_path
        self.static_path = analysis_input.static_path
        self.output_path = analysis_input.output_path
        self.hazard_names = analysis_input.hazard_names
        self.origins_destinations = analysis_input.origins_destinations

    @staticmethod
    def extract_od_nodes_from_graph(
        graph: nx.MultiGraph,
    ) -> list[tuple[str, str]]:
        """
        Extracts all Origin - Destination nodes from the graph, prevents from entries
        with list of nodes for a node.

        Args:
            graph (nx.MultiGraph): Graph containing origin-destination nodes.

        Returns:
            list[tuple[str, str]]]: List containing tuples of origin - destination node combinations.
        """
        _od_nodes = []
        for n, v in graph.nodes(data=True):
            if "od_id" not in v:
                continue
            _o_node_list = list(map(lambda x: (n, x), v["od_id"].split(",")))
            _od_nodes.extend(_o_node_list)
        return _od_nodes

    @staticmethod
    def find_route_ods(
        graph: nx.MultiGraph,
        od_nodes: list[tuple[tuple[str, str], tuple[str, str]]],
        weighing: WeighingEnum,
    ) -> GeoDataFrame:
        # create the routes between all OD pairs
        (
            o_node_list,
            d_node_list,
            origin_list,
            destination_list,
            opt_path_list,
            weighing_list,
            match_ids_list,
            geometries_list,
        ) = ([], [], [], [], [], [], [], [])
        for o, d in tqdm(od_nodes, desc="Finding optimal routes."):
            if nx.has_path(graph, o[0], d[0]):
                # calculate the length of the preferred route and preferred route nodes
                [pref_route, pref_nodes] = nx.single_source_dijkstra(
                    graph, o[0], d[0], weight=weighing.config_value
                )

                # found out which edges belong to the preferred path
                edgesinpath = list(zip(pref_nodes[0:], pref_nodes[1:]))

                pref_edges = []
                match_list = []
                for u, v in edgesinpath:
                    # get edge with the lowest weighing if there are multiple edges that connect u and v
                    _uv_graph = graph[u][v]
                    edge_key = sorted(
                        _uv_graph,
                        key=lambda x, _fgraph=_uv_graph: _fgraph[x][
                            weighing.config_value
                        ],
                    )[0]
                    _uv_graph_edge = _uv_graph[edge_key]
                    if "geometry" in _uv_graph_edge:
                        pref_edges.append(_uv_graph_edge["geometry"])
                    else:
                        pref_edges.append(
                            LineString(
                                [graph.nodes[u]["geometry"], graph.nodes[v]["geometry"]]
                            )
                        )
                    if "rfid" in _uv_graph_edge:
                        match_list.append(_uv_graph_edge["rfid"])

                combined_pref_edges = MultiLineString([])
                for geometry in pref_edges:
                    combined_pref_edges = combined_pref_edges.union(geometry)

                if not combined_pref_edges.is_valid:
                    print(combined_pref_edges.is_valid)
                    print(o[0], d[0])

                # save all data to lists (of lists)
                o_node_list.append(o[0])
                d_node_list.append(d[0])
                origin_list.append(o[1])
                destination_list.append(d[1])
                opt_path_list.append(pref_nodes)
                weighing_list.append(pref_route)
                match_ids_list.append(match_list)
                geometries_list.append(combined_pref_edges)
                # geometries_list.append(pref_edges)

        # Geodataframe to save all the optimal routes
        pref_routes = GeoDataFrame(
            {
                "o_node": o_node_list,
                "d_node": d_node_list,
                "origin": origin_list,
                "destination": destination_list,
                "opt_path": opt_path_list,
                weighing.config_value: weighing_list,
                "match_ids": match_ids_list,
                "geometry": geometries_list,
            },
            geometry="geometry",
            crs="epsg:4326",
        )
        # Remove potential duplicates (o, d node) with a different Origin name.
        _duplicate_columns = ["o_node", "d_node", "destination", "length", "geometry"]
        pref_routes = pref_routes.drop_duplicates(
            subset=_duplicate_columns, keep="first"
        ).reset_index(drop=True)
        return pref_routes

    def _get_origin_destination_pairs(
        self, graph: nx.MultiGraph
    ) -> list[tuple[tuple[str, str], tuple[str, str]]]:
        od_path = self.static_path.joinpath(
            "output_graph", "origin_destination_table.feather"
        )
        od = read_feather(od_path)
        od_pairs = [
            (a, b)
            for a in od.loc[od["o_id"].notnull(), "o_id"]
            for b in od.loc[od["d_id"].notnull(), "d_id"]
        ]
        all_nodes = self.extract_od_nodes_from_graph(graph)
        od_nodes = []
        for aa, bb in od_pairs:
            # it is possible that there are multiple origins/destinations at the same 'entry-point' in the road
            od_nodes.append(
                (
                    [
                        (n, n_name)
                        for n, n_name in all_nodes
                        if (n_name == aa) | (aa in n_name)
                    ][0],
                    [
                        (n, n_name)
                        for n, n_name in all_nodes
                        if (n_name == bb) | (bb in n_name)
                    ][0],
                )
            )
        return od_nodes

    def optimal_route_origin_destination(
        self, graph: nx.MultiGraph, analysis: AnalysisSectionLosses
    ) -> GeoDataFrame:
        # create list of origin-destination pairs
        od_nodes = self._get_origin_destination_pairs(graph)
        pref_routes = self.find_route_ods(graph, od_nodes, analysis.weighing)
        return pref_routes

    def optimal_route_od_link(
        self,
        road_network: GeoDataFrame,
        od_table: GeoDataFrame,
        equity: pd.DataFrame,
    ) -> pd.DataFrame:
        return TrafficAnalysisFactory.get_analysis(
            road_network,
            od_table,
            self.origins_destinations.destinations_names,
            equity,
        ).optimal_route_od_link()

    def execute(self) -> GeoDataFrame:
        _output_path = self.output_path.joinpath(self.analysis.analysis.config_value)

        gdf = self.optimal_route_origin_destination(
            self.graph_file.get_graph(), self.analysis
        )

        if self.analysis.save_traffic and hasattr(
            self.origins_destinations, "origin_count"
        ):
            od_table = read_feather(
                self.static_path.joinpath(
                    "output_graph", "origin_destination_table.feather"
                )
            )
            _equity_weights_file = None
            if self.analysis.equity_weight:
                _equity_weights_file = self.static_path.joinpath(
                    "network", self.analysis.equity_weight
                )
            route_traffic_df = self.optimal_route_od_link(
                gdf,
                od_table,
                TrafficAnalysisFactory.read_equity_weights(_equity_weights_file),
            )
            impact_csv_path = _output_path.joinpath(
                (self.analysis.name.replace(" ", "_") + "_link_traffic.csv"),
            )
            route_traffic_df.to_csv(impact_csv_path, index=False)

        return gdf



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007

    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import logging
from pathlib import Path

import numpy as np
import pandas as pd
from geopandas import GeoDataFrame

from ra2ce.analysis.analysis_config_data.enums.damage_curve_enum import DamageCurveEnum
from ra2ce.analysis.analysis_config_data.enums.risk_calculation_mode_enum import (
    RiskCalculationModeEnum,
)
from ra2ce.analysis.damages.damage_calculation.damage_network_base import (
    DamageNetworkBase,
)
from ra2ce.analysis.damages.shape_to_integrate_object.to_integrate_shaper_factory import (
    ToIntegrateShaperFactory,
)


class DamageNetworkReturnPeriods(DamageNetworkBase):
    """
    A road network gdf with Return-Period based hazard data stored in it,
    and for which damages and risk can be calculated
    @Author: Kees van Ginkel

    Mandatory attributes:
        *self.return_periods* (set of strings): 'e.g
        *self.stats* (set)   : the available statistics
    """

    def __init__(
        self,
        road_gdf: GeoDataFrame,
        val_cols: list[str],
        representative_damage_percentage: float,
    ):
        # Construct using the parent class __init__
        super().__init__(road_gdf, val_cols, representative_damage_percentage)

        self.return_periods = set(
            [x.split("_")[1] for x in val_cols]
        )  # set of unique return_periods

        if not any(self.return_periods):
            raise ValueError("No return_period cols present in hazard data")

    @classmethod
    def construct_from_csv(
        cls, path: Path, representative_damage_percentage: float, sep: str = ";"
    ):
        road_gdf = pd.read_csv(path, sep=sep)
        val_cols = [
            c for c in road_gdf.columns if c.startswith("F_")
        ]  # Find everything starting with 'F'
        return cls(road_gdf, val_cols, representative_damage_percentage)

    ### Controlers for return period based damage and risk calculations
    def main(self, damage_function: DamageCurveEnum, manual_damage_functions):
        """
        Control the damage calculation per return period

        """

        assert len(self.return_periods) > 0, "no return periods identified"
        assert "me" in self.stats, "mean water depth (key: me) is missing"
        assert "fr" in self.stats, "inundated fraction (key: fr) is missing"

        self.do_cleanup_and_mask_creation()

        if damage_function == DamageCurveEnum.HZ:
            self.calculate_damage_HZ(events=self.return_periods)

        if damage_function == DamageCurveEnum.OSD:
            self.calculate_damage_OSdaMage(events=self.return_periods)

        if damage_function == DamageCurveEnum.MAN:
            self.calculate_damage_manual_functions(
                events=self.return_periods,
                manual_damage_functions=manual_damage_functions,
            )

    def control_risk_calculation(
        self,
        damage_function: DamageCurveEnum,
        mode: RiskCalculationModeEnum,
        year: int,
    ):
        """
        Controller of the risk calculation, which calls the correct risk (integration) functions

        Arguments:
            *damage_function* (DamageCurveEnum) : defines the damage estimation method
            *mode* (RiskCalculationModeEnum) : the sort of risk calculation that you want to do, can be:
                                ‘default’, 'cut_from_YYYY_year’, ‘triangle_to_null_YYYY_year’
            *year* (int) : the cutoff year/return period of the risk calculation
            :param damage_function:
        """
        self.verify_damage_data_for_risk_calculation()

        # prepare the parameters of the risk calculation
        to_integrate_shaper = ToIntegrateShaperFactory.get_shaper(
            gdf=self.gdf, damage_function=damage_function
        )
        return_periods = to_integrate_shaper.get_return_periods()
        _to_integrate_dict = to_integrate_shaper.shape_to_integrate_object(
            return_periods=return_periods
        )

        for vulnerability_curve_name, _to_integrate in _to_integrate_dict.items():
            if mode == RiskCalculationModeEnum.DEFAULT:
                _to_integrate = self.rework_damage_data_default(_to_integrate)
                _risk = self.integrate_df_trapezoidal(_to_integrate.copy())

            _rps = list(_to_integrate.columns)

            if mode == RiskCalculationModeEnum.CUT_FROM_YEAR:
                """
                In this mode, the integration mimics the presence of a flood protection
                """

                if year <= min(_rps):
                    raise ValueError(
                        """
                    RA2CE cannot calculate risk in 'cut_from' mode if 
                    Return period of the cutoff ({}) <= smallest available return period ({})
                    Use 'default' mode or 'triangle_to_null_mode' instead.
                                        """.format(
                            year, min(_rps)
                        )
                    )

                elif (
                    min(_rps) < year < max(_rps)
                ):  # if protection level is between min and max RP
                    _to_integrate = self.rework_damage_data_cut_from(
                        _to_integrate, year
                    )
                    _risk = self.integrate_df_trapezoidal(_to_integrate.copy())

                elif year >= max(
                    _rps
                ):  # cutoff is larger or equal than largest return period
                    # risk is return frequency of cutoff
                    # times the damage of the most extreme event
                    _to_integrate = _to_integrate.fillna(0)
                    _risk = _to_integrate[max(_to_integrate.columns)] / year

            elif mode == RiskCalculationModeEnum.TRIANGLE_TO_NULL_YEAR:
                """
                In this mode, an extra data point with zero damage is added at some distance from the smallest known RP,
                and the area of the Triangle this creates is also calculated
                """

                if year >= min(_rps) and year != 0:
                    raise ValueError(
                        """
                    RA2CE cannot calculate risk in 'triangle_to_null' mode if 
                    Return period of the triangle ({}) >= smallest available return period ({})
                    Use 'default' mode or 'cut_from' instead.
                                        """.format(
                            year, min(_rps)
                        )
                    )

                if year == 0:
                    logging.warning(
                        "Available lane data cannot simply be converted to float/int, RA2CE will try a clean-up."
                    )
                    year = 1

                _to_integrate = self.rework_damage_data_triangle_to_null(
                    _to_integrate, year
                )
                _risk = self.integrate_df_trapezoidal(_to_integrate.copy())

            self.gdf[f"risk_{vulnerability_curve_name}"] = _risk

    def verify_damage_data_for_risk_calculation(self):
        """
        Do some data quality and requirement checks before starting the risk calculation

        :return:
        """
        # Check if there is only one unique damage function
        # RP should in the column name
        pass

    @staticmethod
    def integrate_df_trapezoidal(df: pd.DataFrame) -> np.array:
        """
        Arguments:
            *df* (pd.DataFrame) :
                Column names should contain return periods (years)
                Each row should contain a set of damages for one object

        Returns:
            np.array : integrated result per row

        """
        # convert return periods to frequencies
        df.columns = [1 / RP for RP in df.columns]
        # sort columns by ascending frequency
        df = df.sort_index(axis="columns")
        values = df.values
        frequencies = df.columns
        return np.trapz(values, frequencies, axis=1)

    @staticmethod
    def rework_damage_data_default(to_integrate: pd.DataFrame) -> pd.DataFrame:
        """
        Rework the damage data to make it suitable for integration (risk calculation) in default mode

        :param _to_integrate:
        :return: _to_integrate
        """
        # Copy the maximum return period with an infinitely high damage
        _max_return_period = max(to_integrate.columns)
        to_integrate[float("inf")] = to_integrate[_max_return_period]

        # Stop integrating at the last known return period, so no further manipulation needed
        _min_return_period = min(to_integrate.columns)

        to_integrate = to_integrate.fillna(0)

        logging.info(
            """Risk calculation runs in 'default' mode. 
                    Assumptions:
                        - for all return periods > max RP{}, damage = dam_RP{}
                        - for all return periods < min RP{}, damage = 0

                    """.format(
                _max_return_period, _max_return_period, _min_return_period
            )
        )
        return to_integrate

    @staticmethod
    def rework_damage_data_cut_from(
        to_integrate: pd.DataFrame, cutoff_rp: int
    ) -> pd.DataFrame:
        """
        Rework the damage data to make it suitable for integration (risk calculation) in default mode

        :param _to_integrate:
                _cutoff_rp : the return period at which to make the cutoff, aka the flood protection level
        :return: _to_integrate
        """
        _rps = list(to_integrate.columns)

        if cutoff_rp in _rps:
            _dropcols = [rp for rp in _rps if rp < cutoff_rp]
            to_integrate = to_integrate.drop(columns=_dropcols)
        else:
            # find position of first RP value < PL
            # pos = _rps.index(next(i for i in _rps if i < _cutoff_rp))
            # _to_integrate = _to_integrate[_rps[0:pos+1]] #remove all the values with smaller RPs than the PL

            _frequencies = to_integrate.copy()
            _frequencies.columns = [1 / c for c in _frequencies.columns]

            _frequencies[1 / cutoff_rp] = np.nan
            _frequencies = _frequencies.interpolate(method="index", axis=1)

            # Drop the columns outside the cutoff
            _dropcols = [c for c in _frequencies.columns if c > 1 / cutoff_rp]
            _frequencies = _frequencies.drop(columns=_dropcols)

            to_integrate = _frequencies
            to_integrate.columns = [1 / c for c in to_integrate.columns]

            # Copy the maximum return period with an infinitely high damage
            _max_return_period = max(to_integrate.columns)
            to_integrate[float("inf")] = to_integrate[_max_return_period]

            to_integrate = to_integrate.fillna(0)

        logging.info(
            """Risk calculation runs in 'cut_from' mode. 
                                Assumptions:
                                    - for all return periods > max RP{}, damage = dam_RP{}
                                    - damage at cutoff is linearly interpolated from known damages
                                    - no damage for al RPs > RP_cutoff ({})

                                """.format(
                _max_return_period, _max_return_period, cutoff_rp
            )
        )

        return to_integrate

    @staticmethod
    def rework_damage_data_triangle_to_null(
        to_integrate: pd.DataFrame, triangle_end: float
    ) -> pd.DataFrame:
        """
        Rework the damage data to make it suitable for integration (risk calculation) in default mode

        :param _to_integrate:
                _triangle_end_rp : the return period at to end the triangle
        :return: _to_integrate
        """
        # Copy the maximum return period with an infinitely high damage
        _max_return_period = max(to_integrate.columns)
        to_integrate[float("inf")] = to_integrate[_max_return_period]

        # At the return period of the triangle end, set all damage values to zero
        to_integrate[triangle_end] = 0

        to_integrate = to_integrate.sort_index(
            axis="columns", ascending=False
        )  # from large to small RP

        to_integrate = to_integrate.fillna(0)

        logging.info(
            """Risk calculation runs in 'triangle to null' mode. 
                                Assumptions:
                                    - for all return periods > max RP{}, damage = dam_RP{}
                                    - at the end of the triangle {}, damage = 0

                                """.format(
                _max_return_period, _max_return_period, triangle_end
            )
        )
        return to_integrate



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007

    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from dataclasses import dataclass, field
from pathlib import Path
from typing import Callable

from geopandas import GeoDataFrame, read_file, sjoin
from networkx import Graph
from numpy import nanmean

from ra2ce.network.hazard.hazard_intersect.hazard_intersect_builder_base import (
    HazardIntersectBuilderBase,
)


@dataclass
class HazardIntersectBuilderForGpkg(HazardIntersectBuilderBase):
    hazard_field_name: str = ""
    hazard_aggregate_wl: str = ""
    ra2ce_names: list[str] = field(default_factory=list)
    hazard_gpkg_files: list[Path] = field(default_factory=list)

    def _from_networkx(self, hazard_overlay: Graph) -> Graph:
        """Overlays the hazard `gpkg` file over the road segments NetworkX graph.

        Args:
            hazard_overlay (NetworkX graph): The graph that should be overlayed with the hazard `gpkg` file(s).

        Returns:
            hazard_overlay (NetworkX graph): The graph with hazard shapefile(s) data joined
        """

        # TODO check if the CRS of the graph and shapefile match
        def networkx_overlay(hazard_shp_file: Path, race_name: str):
            gdf = read_file(str(hazard_shp_file))
            spatial_index = gdf.sindex

            for u, v, k, edata in hazard_overlay.edges.data(keys=True):
                if "geometry" in edata:
                    possible_matches_index = list(
                        spatial_index.intersection(edata["geometry"].bounds)
                    )
                    possible_matches = gdf.iloc[possible_matches_index]
                    precise_matches = possible_matches[
                        possible_matches.intersects(edata["geometry"])
                    ]

                    if not precise_matches.empty:
                        if self.hazard_aggregate_wl == "max":
                            hazard_overlay[u][v][k][
                                race_name + "_" + self.hazard_aggregate_wl[:2]
                            ] = precise_matches[self.hazard_field_name].max()
                        if self.hazard_aggregate_wl == "min":
                            hazard_overlay[u][v][k][
                                race_name + "_" + self.hazard_aggregate_wl[:2]
                            ] = precise_matches[self.hazard_field_name].min()
                        if self.hazard_aggregate_wl == "mean":
                            hazard_overlay[u][v][k][
                                race_name + "_" + self.hazard_aggregate_wl[:2]
                            ] = nanmean(precise_matches[self.hazard_field_name])
                    else:
                        hazard_overlay[u][v][k][
                            race_name + "_" + self.hazard_aggregate_wl[:2]
                        ] = 0
                else:
                    hazard_overlay[u][v][k][
                        race_name + "_" + self.hazard_aggregate_wl[:2]
                    ] = 0

        self._overlay_hazard_files(networkx_overlay)

        return hazard_overlay

    def _from_geodataframe(self, hazard_overlay: GeoDataFrame) -> GeoDataFrame:
        """Overlays the hazard shapefile over the road segments GeoDataFrame. The gdf is reprojected to the hazard shapefile if necessary.

        Args:
            hazard_overlay (GeoDataFrame): the network geodataframe that should be overlayed with the hazard shapefile(s)

        Returns:
            hazard_overlay (GeoDataFrame): the network geodataframe with hazard shapefile(s) data joined
        """
        gdf_crs_original = hazard_overlay.crs

        def geodataframe_overlay(hazard_shp_file: Path, ra2ce_name: str):
            gdf_hazard = read_file(str(hazard_shp_file))

            if hazard_overlay.crs != gdf_hazard.crs:
                hazard_overlay = hazard_overlay.to_crs(gdf_hazard.crs)

            hazard_overlay = sjoin(
                hazard_overlay,
                gdf_hazard[[self.hazard_field_name, "geometry"]],
                how="left",
            )
            hazard_overlay.rename(
                columns={
                    self.hazard_field_name: ra2ce_name
                    + "_"
                    + self.hazard_aggregate_wl[:2]
                },
                inplace=True,
            )

        self._overlay_hazard_files(geodataframe_overlay)

        if hazard_overlay.crs != gdf_crs_original:
            hazard_overlay = hazard_overlay.to_crs(gdf_crs_original)

        return hazard_overlay

    def _overlay_hazard_files(self, overlay_func: Callable[[str, str], None]):
        for i, _ra2ce_name in enumerate(self.ra2ce_names):
            overlay_func(self.hazard_gpkg_files[i], _ra2ce_name)



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007

    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from dataclasses import dataclass

import geopandas as gpd

from ra2ce.analysis.analysis_protocol import AnalysisProtocol


@dataclass
class AnalysisResultWrapper:
    analysis_result: gpd.GeoDataFrame
    analysis: AnalysisProtocol

    def is_valid_result(self) -> bool:
        """
        Validates whether the `analysis_result` in this wrapper is valid.

        Returns:
            bool: validity of `analysis_result`.
        """
        return (
            isinstance(self.analysis_result, gpd.GeoDataFrame)
            and not self.analysis_result.empty
        )



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007
    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
from pathlib import Path

import pandas as pd

from ra2ce.common.io.readers.file_reader_protocol import FileReaderProtocol
from ra2ce.network.avg_speed.avg_speed import AvgSpeed


class AvgSpeedReader(FileReaderProtocol):
    def read(self, file_path: Path) -> AvgSpeed:
        _avg_speed_data = pd.read_csv(file_path)
        _avg_speed = AvgSpeed()
        for _, row in _avg_speed_data.iterrows():
            _avg_speed.set_avg_speed(
                AvgSpeed.get_road_type_list(row["road_types"]), row["avg_speed"]
            )
        return _avg_speed



from __future__ import annotations

from dataclasses import dataclass, field
from pathlib import Path

from geopandas import GeoDataFrame
from networkx import MultiGraph

from ra2ce.network.graph_files.graph_file import GraphFile
from ra2ce.network.graph_files.graph_files_protocol import GraphFileProtocol
from ra2ce.network.graph_files.network_file import NetworkFile


@dataclass
class GraphFilesCollection:
    """
    Class containing a collection of the graphs and the paths to the files.
    The names of the graph file are assumed to be standardized (e.g. "base_graph.p").
    """

    base_graph: GraphFile = field(
        default_factory=lambda: GraphFile(name="base_graph.p")
    )
    base_graph_hazard: GraphFile = field(
        default_factory=lambda: GraphFile(name="base_graph_hazard.p")
    )
    origins_destinations_graph: GraphFile = field(
        default_factory=lambda: GraphFile(name="origins_destinations_graph.p")
    )
    origins_destinations_graph_hazard: GraphFile = field(
        default_factory=lambda: GraphFile(name="origins_destinations_graph_hazard.p")
    )
    base_network: NetworkFile = field(
        default_factory=lambda: NetworkFile(name="base_network.feather")
    )
    base_network_hazard: NetworkFile = field(
        default_factory=lambda: NetworkFile(name="base_network_hazard.feather")
    )
    locations_hazard: NetworkFile = field(
        default_factory=lambda: NetworkFile(name="locations_hazard.feather")
    )

    @property
    def _graph_collection(self) -> list[GraphFile | NetworkFile]:
        return [
            self.base_graph,
            self.base_graph_hazard,
            self.origins_destinations_graph,
            self.origins_destinations_graph_hazard,
            self.base_network,
            self.base_network_hazard,
            self.locations_hazard,
        ]

    def has_graphs(self) -> bool:
        """
        Tests if any of the types has a graph already read

        Returns:
            bool: True if any graph already read
        """

        def has_graph(gf: GraphFileProtocol) -> bool:
            return gf.graph is not None

        return any(filter(has_graph, self._graph_collection))

    def _get_graph_file(self, graph_file_type: str) -> GraphFileProtocol:
        """
        Get the graph (object)

        Args:
            graph_file_type (str): Type of graph file

        Raises:
            ValueError: If the graph_file_type is not one of the known types

        Returns:
            GraphFileProtocol: Graph object of that specific graph_file_type
        """
        _gf = next(
            (
                gf
                for gf in self._graph_collection
                if Path(gf.name).stem == graph_file_type
            ),
            None,
        )
        if _gf is None:
            raise ValueError(f"Unknown graph file type {graph_file_type} provided.")
        return _gf

    def get_graph(self, graph_file_type: str) -> MultiGraph | GeoDataFrame:
        """
        Get the graph

        Args:
            graph_file_type (str): Type of graph file

        Returns:
            GraphFileProtocol: Graph of that specific graph_file_type
        """
        _gf = self._get_graph_file(graph_file_type)
        return _gf.get_graph()

    def get_file(self, graph_file_type: str) -> Path | None:
        """
        Get the file path to the graph

        Args:
            graph_file_type (str): Type of graph file

        Returns:
            Path: Path to the graph file
        """
        _gf = self._get_graph_file(graph_file_type)
        return _gf.file

    @classmethod
    def set_files(cls, parent_dir: Path) -> GraphFilesCollection:
        """
        Create a new collection with 1 or more graph files that match the default names

        Args:
            parent_dir (Path): Path of the parent folder in which the files are searched

        Returns:
            GraphFilesCollection: Collection of graphs
        """
        _collection = cls()

        _collection.base_graph.read_graph(parent_dir)
        _collection.base_graph_hazard.read_graph(parent_dir)
        _collection.origins_destinations_graph.read_graph(parent_dir)
        _collection.origins_destinations_graph_hazard.read_graph(parent_dir)
        _collection.base_network.read_graph(parent_dir)
        _collection.base_network_hazard.read_graph(parent_dir)
        _collection.locations_hazard.read_graph(parent_dir)

        return _collection

    def set_file(self, file: Path) -> None:
        """
        Set a path to a graph via the collection

        Args:
            file (Path): Path of the graph

        Raises:
            ValueError: If the graph_file_type is not one of the known types
        """
        _gf = self._get_graph_file(file.stem)
        _gf.folder = file.parent

    def set_graph(self, graph_file_type: str, graph: MultiGraph | GeoDataFrame):
        """
        Set a graph via the collection based on type of graph

        Args:
            graph_file_type (str): Type of graph file
            graph (MultiGraph | GeoDataFrame): The graph

        Raises:
            ValueError: If the graph_file_type is not one of the known types
        """
        _gf = self._get_graph_file(graph_file_type)
        _gf.graph = graph

    def read_graph(self, file: Path) -> None:
        """
        Read a graph file via the collection

        Args:
            file (Path): Path fo the graph

        Raises:
            ValueError: If the graph_file_type is not one of the known types

        Returns: None
        """
        _gf = next((gf for gf in self._graph_collection if gf.name == file.name), None)
        if _gf is None:
            raise ValueError(f"Unknown graph file {file} provided.")
        _gf.read_graph(file.parent)



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007

    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import geopandas as gpd
import numpy as np
import pandas as pd

from ra2ce.analysis.losses.traffic_analysis.accumulated_traffic_dataclass import (
    AccumulatedTraffic,
)
from ra2ce.analysis.losses.traffic_analysis.traffic_analysis_base import (
    TrafficAnalysisBase,
)


class EquityAnalysis(TrafficAnalysisBase):
    """
    Specialization from `TrafficAnalysisBase` which takes into account the equity data (weights per region).
    """

    road_network: gpd.GeoDataFrame
    od_table: gpd.GeoDataFrame
    equity_data: pd.DataFrame
    destinations_names: str

    def __init__(
        self,
        road_network: gpd.GeoDataFrame,
        od_table: gpd.GeoDataFrame,
        destination_names: str,
        equity_data: pd.DataFrame,
    ) -> None:
        """
        Args:
            road_network (gpd.GeoDataFrame): Geodataframe containing the overall network information.
            od_table (gpd.GeoDataFrame): GeoDataFrame representing the Origins - Destinations table.
            destination_names (str): Destination nodes.
            equity_data ( pd.DataFrame): Pandas DataFrame representing the equity traffic data.
        """
        self.road_network = road_network
        self.od_table = od_table
        self.destinations_names = destination_names
        self.equity_data = equity_data
        self._set_values_prioritarian(self.equity_data, self.od_table)

    def _set_values_prioritarian(
        self, equity_data: pd.DataFrame, od_table_data: gpd.GeoDataFrame
    ) -> np.array:
        prioritarian_mapping = dict(zip(equity_data["region"], equity_data["weight"]))
        prioritarian_mapping.update({"Not assigned": 1})
        self.od_table["values_prioritarian"] = (
            od_table_data["region"].map(prioritarian_mapping) * od_table_data["values"]
        )

    def _get_accumulated_traffic_from_node(
        self, o_node: str, total_d_nodes: int
    ) -> AccumulatedTraffic:
        _accumulated_traffic = AccumulatedTraffic(egalitarian=1)
        _accumulated_traffic.utilitarian = self._get_recorded_traffic_in_node(
            o_node, "values", total_d_nodes
        )
        _accumulated_traffic.prioritarian = self._get_recorded_traffic_in_node(
            o_node, "values_prioritarian", total_d_nodes
        )
        return _accumulated_traffic

    def _get_route_traffic(
        self, traffic_data: dict[str, AccumulatedTraffic]
    ) -> pd.DataFrame:
        def convert_to_df(dict_item: tuple[str, AccumulatedTraffic]):
            node_id, traffic_values = dict_item
            u_node, v_node = self._get_key_nodes(node_id)
            return (
                u_node,
                v_node,
                traffic_values.utilitarian,
                traffic_values.egalitarian,
                traffic_values.prioritarian,
            )

        return pd.DataFrame(
            map(convert_to_df, traffic_data.items()),
            columns=[
                "u",
                "v",
                "traffic",
                "traffic_egalitarian",
                "traffic_prioritarian",
            ],
        )



from __future__ import annotations

from ra2ce.configuration.ra2ce_enum_base import Ra2ceEnumBase


class WeighingEnum(Ra2ceEnumBase):
    NONE = 0
    LENGTH = 1
    TIME = 2
    INVALID = 99

    @classmethod
    def get_enum(cls, input: str | None) -> WeighingEnum:
        return super().get_enum(input)



import logging
from pathlib import Path

from geopandas import GeoDataFrame

from ra2ce.analysis.analysis_config_data.analysis_config_data import (
    AnalysisSectionDamages,
)
from ra2ce.analysis.analysis_config_data.enums.damage_curve_enum import DamageCurveEnum
from ra2ce.analysis.analysis_config_data.enums.event_type_enum import EventTypeEnum
from ra2ce.analysis.analysis_config_data.enums.risk_calculation_mode_enum import (
    RiskCalculationModeEnum,
)
from ra2ce.analysis.analysis_input_wrapper import AnalysisInputWrapper
from ra2ce.analysis.damages.analysis_damages_protocol import AnalysisDamagesProtocol
from ra2ce.analysis.damages.damage.manual_damage_functions import ManualDamageFunctions
from ra2ce.analysis.damages.damage_calculation import (
    DamageNetworkEvents,
    DamageNetworkReturnPeriods,
)
from ra2ce.network.graph_files.network_file import NetworkFile


class Damages(AnalysisDamagesProtocol):
    analysis: AnalysisSectionDamages
    graph_file: NetworkFile
    graph_file_hazard: NetworkFile
    input_path: Path
    output_path: Path

    def __init__(
        self,
        analysis_input: AnalysisInputWrapper,
    ) -> None:
        self.analysis = analysis_input.analysis
        self.graph_file = None
        self.graph_file_hazard = analysis_input.graph_file_hazard
        self.input_path = analysis_input.input_path
        self.output_path = analysis_input.output_path

    def execute(self) -> GeoDataFrame:
        def _rename_road_gdf_to_conventions(road_gdf_columns: list[str]) -> list[str]:
            """
            Rename the columns in the road_gdf to the conventions of the ra2ce documentation

            'eg' RP100_fr -> F_RP100_me
                        -> F_EV1_mi

            """
            cs = road_gdf_columns
            ### Handle return period columns
            new_cols = []
            for c in cs:
                if c.startswith("RP") or c.startswith("EV"):
                    new_cols.append(f"{hazard_prefix}_" + c)
                else:
                    new_cols.append(c)

            ### Todo add handling of events if this gives a problem
            return new_cols

        hazard_prefix = "F"
        # Open the network with hazard data
        road_gdf = self.graph_file_hazard.get_graph()
        road_gdf.columns = _rename_road_gdf_to_conventions(road_gdf.columns)

        # Find the hazard columns; these may be events or return periods
        val_cols = [
            col for col in road_gdf.columns if f"{hazard_prefix}" in col.split("_")
        ]

        # Read the desired damage function
        damage_function = self.analysis.damage_curve

        # If you want to use manual damage functions, these need to be loaded first
        manual_damage_functions = None
        if self.analysis.damage_curve == DamageCurveEnum.MAN:
            manual_damage_functions = ManualDamageFunctions()
            manual_damage_functions.find_damage_functions(
                folder=(self.input_path.joinpath("damage_functions"))
            )
            manual_damage_functions.load_damage_functions()

        # Choose between event or return period based analysis
        if self.analysis.event_type == EventTypeEnum.EVENT:
            event_gdf = DamageNetworkEvents(
                road_gdf, val_cols, self.analysis.representative_damage_percentage
            )
            event_gdf.main(
                damage_function=damage_function,
                manual_damage_functions=manual_damage_functions,
            )

            return event_gdf.gdf

        elif self.analysis.event_type == EventTypeEnum.RETURN_PERIOD:
            return_period_gdf = DamageNetworkReturnPeriods(
                road_gdf, val_cols, self.analysis.representative_damage_percentage
            )
            return_period_gdf.main(
                damage_function=damage_function,
                manual_damage_functions=manual_damage_functions,
            )

            if (
                self.analysis.risk_calculation_mode != RiskCalculationModeEnum.INVALID
            ):  # Check if risk_calculation is demanded
                if self.analysis.risk_calculation_mode != RiskCalculationModeEnum.NONE:
                    return_period_gdf.control_risk_calculation(
                        self.analysis.damage_curve,
                        mode=self.analysis.risk_calculation_mode,
                        year=self.analysis.risk_calculation_year,
                    )

            else:
                logging.info(
                    """No parameters for risk calculation are specified. 
                             Add key [risk_calculation_mode] to analyses.ini."""
                )

            return return_period_gdf.gdf

        raise ValueError(
            "The hazard calculation does not know what to do if the analysis specifies {}".format(
                self.analysis.event_type
            )
        )



import logging
from pathlib import Path

from geopandas import GeoDataFrame

from ra2ce.analysis.analysis_config_data.analysis_config_data import (
    AnalysisSectionLosses,
)
from ra2ce.analysis.analysis_input_wrapper import AnalysisInputWrapper
from ra2ce.analysis.losses.analysis_losses_protocol import AnalysisLossesProtocol
from ra2ce.analysis.losses.origin_closest_destination import OriginClosestDestination
from ra2ce.network.graph_files.graph_file import GraphFile
from ra2ce.network.hazard.hazard_names import HazardNames
from ra2ce.network.network_config_data.network_config_data import (
    OriginsDestinationsSection,
)
from ra2ce.network.networks_utils import graph_to_gpkg


class OptimalRouteOriginClosestDestination(AnalysisLossesProtocol):
    analysis: AnalysisSectionLosses
    graph_file_hazard: GraphFile
    input_path: Path
    static_path: Path
    output_path: Path
    hazard_names: HazardNames
    origins_destinations: OriginsDestinationsSection
    file_id: str
    _analysis_input: AnalysisInputWrapper

    def __init__(
        self,
        analysis_input: AnalysisInputWrapper,
    ) -> None:
        self.analysis = analysis_input.analysis
        self.graph_file_hazard = analysis_input.graph_file_hazard
        self.input_path = analysis_input.input_path
        self.static_path = analysis_input.static_path
        self.output_path = analysis_input.output_path
        self.hazard_names = analysis_input.hazard_names
        self.origins_destinations = analysis_input.origins_destinations
        self.file_id = analysis_input.file_id
        self._analysis_input = analysis_input

    def _save_gdf(self, gdf: GeoDataFrame, save_path: Path):
        """Takes in a geodataframe object and outputs shapefiles at the paths indicated by edge_shp and node_shp

        Arguments:
            gdf [geodataframe]: geodataframe object to be converted
            save_path [str]: output path including extension for edges shapefile
        Returns:
            None
        """
        # save to shapefile
        gdf.crs = "epsg:4326"  # TODO: decide if this should be variable with e.g. an output_crs configured

        for col in gdf.columns:
            if gdf[col].dtype == object and col != gdf.geometry.name:
                gdf[col] = gdf[col].astype(str)

        if save_path.exists():
            save_path.unlink()
        gdf.to_file(save_path, driver="GPKG")
        logging.info("Results saved to: {}".format(save_path))

    def execute(self) -> GeoDataFrame:
        def _save_gpkg_analysis(
            base_graph,
            to_save_gdf: list[GeoDataFrame],
            to_save_gdf_names: list[str],
        ):
            for to_save, save_name in zip(to_save_gdf, to_save_gdf_names):
                if not to_save.empty:
                    gpkg_path = _output_path.joinpath(
                        self.analysis.name.replace(" ", "_") + f"_{save_name}.gpkg"
                    )
                    self._save_gdf(to_save, gpkg_path)

            # Save the Graph
            gpkg_path_nodes = _output_path.joinpath(
                self.analysis.name.replace(" ", "_") + "_results_nodes.gpkg"
            )
            gpkg_path_edges = _output_path.joinpath(
                self.analysis.name.replace(" ", "_") + "_results_edges.gpkg"
            )
            graph_to_gpkg(base_graph, gpkg_path_edges, gpkg_path_nodes)

        _output_path = self.output_path.joinpath(self.analysis.analysis.config_value)

        analyzer = OriginClosestDestination(self._analysis_input)
        (
            base_graph,
            opt_routes,
            destinations,
        ) = analyzer.optimal_route_origin_closest_destination()
        if self.analysis.save_gpkg:
            # Save the GeoDataFrames
            to_save_gdf = [destinations, opt_routes]
            to_save_gdf_names = ["destinations", "optimal_routes"]
            _save_gpkg_analysis(base_graph, to_save_gdf, to_save_gdf_names)

        if self.analysis.save_csv:
            csv_path = _output_path.joinpath(
                self.analysis.name.replace(" ", "_") + "_destinations.csv"
            )
            del destinations["geometry"]
            destinations.to_csv(csv_path, index=False)

            csv_path = _output_path.joinpath(
                self.analysis.name.replace(" ", "_") + "_optimal_routes.csv"
            )
            del opt_routes["geometry"]
            opt_routes.to_csv(csv_path, index=False)

        return None



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007

    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from typing import Protocol

from geopandas import GeoDataFrame
from networkx import Graph


class HazardIntersectBuilderProtocol(Protocol):
    def get_intersection(
        self, hazard_overlay: GeoDataFrame | Graph
    ) -> GeoDataFrame | Graph:
        """
        Retrieves the resulting network from intersecting the hazard layer with a graph.

        Args:
            hazard_overlay (GeoDataFrame | Graph): Layer containing hazards.

        Returns:
            GeoDataFrame | Graph: Intersected graph with hazards.
        """
        pass



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007

    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from dataclasses import dataclass

from geopandas import GeoDataFrame
from networkx import Graph
from pandas import read_csv

from ra2ce.network.hazard.hazard_intersect.hazard_intersect_builder_base import (
    HazardIntersectBuilderBase,
)
from ra2ce.network.networks_utils import graph_from_gdf, graph_to_gdf


@dataclass
class HazardIntersectBuilderForTable(HazardIntersectBuilderBase):
    hazard_field_name: str = ""
    network_file_id: str = ""
    hazard_id: str = ""
    ra2ce_name_key: str = "RA2CE name"

    def _from_network_x(self, hazard_overlay: Graph) -> Graph:
        """Joins a table with IDs and hazard information with the road segments with corresponding IDs."""
        gdf, gdf_nodes = graph_to_gdf(hazard_overlay, save_nodes=True)
        gdf = self._from_geodataframe(gdf)

        # TODO: Check if the graph is created again correctly.
        hazard_overlay = graph_from_gdf(gdf, gdf_nodes)
        return hazard_overlay

    def _from_geodataframe(self, hazard_overlay: GeoDataFrame):
        """Joins a table with IDs and hazard information with the road segments with corresponding IDs."""
        for haz in self.hazard_files.table:
            if haz.suffix in [".csv"]:
                hazard_overlay = self._join_table(hazard_overlay, haz)
        return hazard_overlay

    def _join_table(self, graph: Graph, hazard: str) -> Graph:
        df = read_csv(hazard)
        df = df[self.hazard_field_name]
        graph = graph.merge(
            df,
            how="left",
            left_on=self.network_file_id,
            right_on=self.hazard_id,
        )

        graph.rename(
            columns={
                self.hazard_field_name: [
                    n[:-3] for n in self.hazard_name_table[self.ra2ce_name_key]
                ][0]
            },
            inplace=True,
        )  # Check if this is the right name
        return graph



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007
    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any

import pandas as pd

from ra2ce.common.io.readers.file_reader_protocol import FileReaderProtocol


class LossesInputDataReaderBase(ABC, FileReaderProtocol):
    """
    Base class for reading losses input data from a csv file.
    """

    csv_columns: list[str] = []
    separator: str = ";"
    object_type: type[Any]

    @abstractmethod
    def _parse_df(self, df: pd.DataFrame) -> Any:
        """
        Abstract method to parse the data from a DataFrame.

        Args:
            df (pd.DataFrame): The parsed data from the csv file.

        Returns:
            Any: Dataclass containing the parsed data.
        """
        pass

    def read(self, file_path: Path | None) -> Any:
        if not file_path or not file_path.exists():
            logging.warning(
                "No csv file found at %s. Using default values for %s.",
                file_path,
                self.object_type,
            )
            return self.object_type()

        _df = pd.read_csv(file_path, sep=self.separator, on_bad_lines="skip")
        if "geometry" in _df.columns:
            raise ValueError(
                f"The csv file in {file_path} should not have a geometry column"
            )

        if not all(col in _df.columns for col in self.csv_columns):
            raise ValueError(
                f"The csv file in {file_path} should have columns {', '.join(self.csv_columns)}",
            )

        return self._parse_df(_df)



from ra2ce.analysis.damages.damage_calculation.damage_network_events import (
    DamageNetworkEvents,
)
from ra2ce.analysis.damages.damage_calculation.damage_network_return_periods import (
    DamageNetworkReturnPeriods,
)



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007

    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import os
from collections import OrderedDict
from pathlib import Path
from typing import Any

import numpy as np
import pandas as pd
from pandas import DataFrame
from scipy.interpolate import interp1d


def dataframe_lookup(row: pd.Series, lookup_df: DataFrame, columns: list) -> Any:
    row_values = [row[column] for column in columns]
    return lookup_df.loc[tuple(row_values)]


class LookUp:
    """ " This namespace contains several lookup tables, used e.g. for road damage calculation."""

    @staticmethod
    def road_mapping():
        """Mapping of OSM road infrastructure types"""

        mapping_dict = {
            "disused": "none",
            "dummy": "none",
            "planned": "none",
            "platform": "none",
            "unsurfaced": "track",
            "traffic_island": "other",
            "razed": "none",
            "abandoned": "none",
            "services": "none",
            "proposed": "none",
            "corridor": "track",
            "bus_guideway": "other",
            "bus_stop": "other",
            "rest_area": "other",
            "yes": "other",
            "trail": "track",
            "escape": "other",
            "raceway": "other",
            "emergency_access_point": "none",
            "emergency_bay": "other",
            "construction": "other",
            "bridleway": "none",
            "cycleway": "other",
            "footway": "track",
            "living_street": "other",
            "path": "track",
            "pedestrian": "other",
            "primary": "primary",
            "primary_link": "primary",
            "residential": "other",
            "road": "other",
            "secondary": "secondary",
            "secondary_link": "secondary",
            "service": "other",
            "steps": "none",
            "tertiary": "tertiary",
            "tertiary_link": "tertiary",
            "track": "track",
            "unclassified": "other",
            "trunk": "trunk",
            "motorway": "motorway",
            "trunk_link": "trunk",
            "motorway_link": "motorway",
            "elevator": "none",
            "access": "none",
            "crossing": "other",
            "mini_roundabout": "other",
            "passing_place": "other",
            "turning_circle": "other",
            "motorway_junction": "motorway",
        }
        return mapping_dict

    @staticmethod
    def road_lanes():
        """Mapping of road lanes, this is used as a default in case of missings. Number of lanes differ per country."""

        return OrderedDict(
            [
                (
                    "AL",
                    OrderedDict(
                        [
                            ("motorway", 2),
                            ("trunk", 2),
                            ("primary", 2),
                            ("secondary", 2),
                            ("tertiary", 1),
                            ("track", 1),
                            ("other", 1),
                            ("none", 1),
                        ]
                    ),
                ),
                (
                    "AT",
                    OrderedDict(
                        [
                            ("motorway", 2),
                            ("trunk", 2),
                            ("primary", 2),
                            ("secondary", 2),
                            ("tertiary", 1),
                            ("track", 1),
                            ("other", 1),
                            ("none", 1),
                        ]
                    ),
                ),
                (
                    "BE",
                    OrderedDict(
                        [
                            ("motorway", 2),
                            ("trunk", 2),
                            ("primary", 2),
                            ("secondary", 2),
                            ("tertiary", 2),
                            ("track", 1),
                            ("other", 1),
                            ("none", 1),
                        ]
                    ),
                ),
                (
                    "BG",
                    OrderedDict(
                        [
                            ("motorway", 2),
                            ("trunk", 2),
                            ("primary", 2),
                            ("secondary", 2),
                            ("tertiary", 2),
                            ("track", 1),
                            ("other", 1),
                            ("none", 1),
                        ]
                    ),
                ),
                (
                    "CH",
                    OrderedDict(
                        [
                            ("motorway", 2),
                            ("trunk", 2),
                            ("primary", 2),
                            ("secondary", 2),
                            ("tertiary", 1),
                            ("track", 1),
                            ("other", 1),
                            ("none", 1),
                        ]
                    ),
                ),
                (
                    "CZ",
                    OrderedDict(
                        [
                            ("motorway", 2),
                            ("trunk", 2),
                            ("primary", 2),
                            ("secondary", 2),
                            ("tertiary", 2),
                            ("track", 1),
                            ("other", 1),
                            ("none", 1),
                        ]
                    ),
                ),
                (
                    "DE",
                    OrderedDict(
                        [
                            ("motorway", 2),
                            ("trunk", 2),
                            ("primary", 2),
                            ("secondary", 2),
                            ("tertiary", 2),
                            ("track", 1),
                            ("other", 1),
                            ("none", 1),
                        ]
                    ),
                ),
                (
                    "DK",
                    OrderedDict(
                        [
                            ("motorway", 2),
                            ("trunk", 2),
                            ("primary", 2),
                            ("secondary", 2),
                            ("tertiary", 2),
                            ("track", 1),
                            ("other", 1),
                            ("none", 1),
                        ]
                    ),
                ),
                (
                    "EE",
                    OrderedDict(
                        [
                            ("motorway", 1),
                            ("trunk", 2),
                            ("primary", 2),
                            ("secondary", 2),
                            ("tertiary", 2),
                            ("track", 1),
                            ("other", 1),
                            ("none", 1),
                        ]
                    ),
                ),
                (
                    "EL",
                    OrderedDict(
                        [
                            ("motorway", 2),
                            ("trunk", 2),
                            ("primary", 2),
                            ("secondary", 2),
                            ("tertiary", 2),
                            ("track", 1),
                            ("other", 1),
                            ("none", 1),
                        ]
                    ),
                ),
                (
                    "ES",
                    OrderedDict(
                        [
                            ("motorway", 2),
                            ("trunk", 2),
                            ("primary", 2),
                            ("secondary", 2),
                            ("tertiary", 2),
                            ("track", 1),
                            ("other", 1),
                            ("none", 1),
                        ]
                    ),
                ),
                (
                    "FI",
                    OrderedDict(
                        [
                            ("motorway", 2),
                            ("trunk", 2),
                            ("primary", 2),
                            ("secondary", 2),
                            ("tertiary", 2),
                            ("track", 1),
                            ("other", 1),
                            ("none", 1),
                        ]
                    ),
                ),
                (
                    "FR",
                    OrderedDict(
                        [
                            ("motorway", 2),
                            ("trunk", 2),
                            ("primary", 2),
                            ("secondary", 2),
                            ("tertiary", 2),
                            ("track", 1),
                            ("other", 1),
                            ("none", 1),
                        ]
                    ),
                ),
                (
                    "HR",
                    OrderedDict(
                        [
                            ("motorway", 2),
                            ("trunk", 2),
                            ("primary", 2),
                            ("secondary", 2),
                            ("tertiary", 2),
                            ("track", 1),
                            ("other", 1),
                            ("none", 1),
                        ]
                    ),
                ),
                (
                    "HU",
                    OrderedDict(
                        [
                            ("motorway", 2),
                            ("trunk", 2),
                            ("primary", 2),
                            ("secondary", 2),
                            ("tertiary", 2),
                            ("track", 1),
                            ("other", 1),
                            ("none", 1),
                        ]
                    ),
                ),
                (
                    "IE",
                    OrderedDict(
                        [
                            ("motorway", 2),
                            ("trunk", 2),
                            ("primary", 2),
                            ("secondary", 2),
                            ("tertiary", 1),
                            ("track", 1),
                            ("other", 1),
                            ("none", 1),
                        ]
                    ),
                ),
                (
                    "IS",
                    OrderedDict(
                        [
                            ("motorway", 2),
                            ("trunk", 2),
                            ("primary", 2),
                            ("secondary", 2),
                            ("tertiary", 1),
                            ("track", 1),
                            ("other", 1),
                            ("none", 1),
                        ]
                    ),
                ),
                (
                    "IT",
                    OrderedDict(
                        [
                            ("motorway", 2),
                            ("trunk", 2),
                            ("primary", 2),
                            ("secondary", 2),
                            ("tertiary", 2),
                            ("track", 1),
                            ("other", 1),
                            ("none", 1),
                        ]
                    ),
                ),
                (
                    "LT",
                    OrderedDict(
                        [
                            ("motorway", 2),
                            ("trunk", 2),
                            ("primary", 2),
                            ("secondary", 2),
                            ("tertiary", 2),
                            ("track", 1),
                            ("other", 1),
                            ("none", 1),
                        ]
                    ),
                ),
                (
                    "LU",
                    OrderedDict(
                        [
                            ("motorway", 2),
                            ("trunk", 3),
                            ("primary", 2),
                            ("secondary", 2),
                            ("tertiary", 2),
                            ("track", 1),
                            ("other", 1),
                            ("none", 1),
                        ]
                    ),
                ),
                (
                    "LV",
                    OrderedDict(
                        [
                            ("motorway", 1),
                            ("trunk", 2),
                            ("primary", 2),
                            ("secondary", 2),
                            ("tertiary", 2),
                            ("track", 1),
                            ("other", 1),
                            ("none", 1),
                        ]
                    ),
                ),
                (
                    "LI",
                    OrderedDict(
                        [
                            ("motorway", 1),
                            ("trunk", 1),
                            ("primary", 2),
                            ("secondary", 2),
                            ("tertiary", 2),
                            ("track", 1),
                            ("other", 1),
                            ("none", 1),
                        ]
                    ),
                ),
                (
                    "ME",
                    OrderedDict(
                        [
                            ("motorway", 1),
                            ("trunk", 1),
                            ("primary", 2),
                            ("secondary", 1),
                            ("tertiary", 2),
                            ("track", 1),
                            ("other", 1),
                            ("none", 1),
                        ]
                    ),
                ),
                (
                    "MK",
                    OrderedDict(
                        [
                            ("motorway", 2),
                            ("trunk", 2),
                            ("primary", 2),
                            ("secondary", 2),
                            ("tertiary", 2),
                            ("track", 1),
                            ("other", 1),
                            ("none", 1),
                        ]
                    ),
                ),
                (
                    "NL",
                    OrderedDict(
                        [
                            ("motorway", 2),
                            ("trunk", 2),
                            ("primary", 2),
                            ("secondary", 2),
                            ("tertiary", 1),
                            ("track", 1),
                            ("other", 1),
                            ("none", 1),
                        ]
                    ),
                ),
                (
                    "NO",
                    OrderedDict(
                        [
                            ("motorway", 2),
                            ("trunk", 2),
                            ("primary", 2),
                            ("secondary", 2),
                            ("tertiary", 1),
                            ("track", 1),
                            ("other", 1),
                            ("none", 1),
                        ]
                    ),
                ),
                (
                    "PL",
                    OrderedDict(
                        [
                            ("motorway", 2),
                            ("trunk", 2),
                            ("primary", 2),
                            ("secondary", 2),
                            ("tertiary", 2),
                            ("track", 1),
                            ("other", 1),
                            ("none", 1),
                        ]
                    ),
                ),
                (
                    "PT",
                    OrderedDict(
                        [
                            ("motorway", 2),
                            ("trunk", 2),
                            ("primary", 2),
                            ("secondary", 2),
                            ("tertiary", 2),
                            ("track", 1),
                            ("other", 1),
                            ("none", 1),
                        ]
                    ),
                ),
                (
                    "RO",
                    OrderedDict(
                        [
                            ("motorway", 2),
                            ("trunk", 2),
                            ("primary", 2),
                            ("secondary", 2),
                            ("tertiary", 2),
                            ("track", 1),
                            ("other", 1),
                            ("none", 1),
                        ]
                    ),
                ),
                (
                    "RS",
                    OrderedDict(
                        [
                            ("motorway", 2),
                            ("trunk", 2),
                            ("primary", 2),
                            ("secondary", 2),
                            ("tertiary", 2),
                            ("track", 1),
                            ("other", 1),
                            ("none", 1),
                        ]
                    ),
                ),
                (
                    "SE",
                    OrderedDict(
                        [
                            ("motorway", 2),
                            ("trunk", 2),
                            ("primary", 2),
                            ("secondary", 2),
                            ("tertiary", 2),
                            ("track", 1),
                            ("other", 1),
                            ("none", 1),
                        ]
                    ),
                ),
                (
                    "SI",
                    OrderedDict(
                        [
                            ("motorway", 2),
                            ("trunk", 2),
                            ("primary", 2),
                            ("secondary", 2),
                            ("tertiary", 2),
                            ("track", 1),
                            ("other", 1),
                            ("none", 1),
                        ]
                    ),
                ),
                (
                    "SK",
                    OrderedDict(
                        [
                            ("motorway", 2),
                            ("trunk", 2),
                            ("primary", 2),
                            ("secondary", 2),
                            ("tertiary", 2),
                            ("track", 1),
                            ("other", 1),
                            ("none", 1),
                        ]
                    ),
                ),
                (
                    "TR",
                    OrderedDict(
                        [
                            ("motorway", 3),
                            ("trunk", 2),
                            ("primary", 2),
                            ("secondary", 2),
                            ("tertiary", 2),
                            ("track", 1),
                            ("other", 1),
                            ("none", 1),
                        ]
                    ),
                ),
                (
                    "UK",
                    OrderedDict(
                        [
                            ("motorway", 3),
                            ("trunk", 2),
                            ("primary", 2),
                            ("secondary", 2),
                            ("tertiary", 2),
                            ("track", 1),
                            ("other", 1),
                            ("none", 1),
                        ]
                    ),
                ),
            ]
        )

    @staticmethod
    def get_lane_number_damage_correction():
        """Lookup table for correction factor of damage for lanes"""
        lookup_dict = OrderedDict(
            [
                ("motorway", {1: 0.75, 2: 1.0, 3: 1.25, 4: 1.5, 5: 1.75, 6: 2.0}),
                ("trunk", {1: 0.75, 2: 1.0, 3: 1.25, 4: 1.5, 5: 1.75, 6: 2.0}),
                ("primary", {1: 0.75, 2: 1.0, 3: 1.25, 4: 1.5, 5: 1.75, 6: 2.0}),
                ("secondary", {1: 0.75, 2: 1.0, 3: 1.25, 4: 1.5, 5: 1.75, 6: 2.0}),
                ("tertiary", {1: 0.75, 2: 1.0, 3: 1.5, 4: 1.75, 5: 2.0, 6: 2.25}),
                ("other", {1: 1.0, 2: 1.25, 3: 1.5, 4: 1.75, 5: 2.0, 6: 2.25}),
                ("track", {1: 1.0, 2: 1.25, 3: 1.5, 4: 1.75, 5: 2.0, 6: 2.25}),
                ("none", {1: 1.0, 2: 1.25, 3: 1.5, 4: 1.75, 5: 2.0, 6: 2.25}),
            ]
        )

        return lookup_dict

    @staticmethod
    def max_damages():
        """Lookup table for max damages"""
        return OrderedDict(
            [
                (
                    "Lower",
                    OrderedDict(
                        [
                            ("motorway", 1750000),
                            ("trunk", 1250000),
                            ("primary", 1000000),
                            ("secondary", 500000),
                            ("tertiary", 200000),
                            ("other", 100000),
                            ("track", 20000),
                            ("none", 0),
                        ]
                    ),
                ),
                (
                    "Upper",
                    OrderedDict(
                        [
                            ("motorway", 17500000),
                            ("trunk", 3750000),
                            ("primary", 3000000),
                            ("secondary", 1500000),
                            ("tertiary", 600000),
                            ("other", 300000),
                            ("track", 50000),
                            ("none", 0),
                        ]
                    ),
                ),
            ]
        )

    @staticmethod
    def get_max_damages_osd():
        """Lookup table for max damages of the OSdaMage damage functions"""

        # Note that these values have been converted to euro/m road length
        return OrderedDict(
            [
                (
                    "Lower",
                    OrderedDict(
                        [
                            ("motorway", 1750),
                            ("trunk", 1250),
                            ("primary", 1000),
                            ("secondary", 500),
                            ("tertiary", 200),
                            ("other", 100),
                            ("track", 20),
                            ("none", 0),
                        ]
                    ),
                ),
                (
                    "Upper",
                    OrderedDict(
                        [
                            ("motorway", 17500),
                            ("trunk", 3750),
                            ("primary", 3000),
                            ("secondary", 1500),
                            ("tertiary", 600),
                            ("other", 300),
                            ("track", 50),
                            ("none", 0),
                        ]
                    ),
                ),
            ]
        )

    @staticmethod
    def get_max_damages_huizinga() -> dict:
        """Lookup table for max damages calculated with huizinga for number of lanes

        Output: dict:
         - road types are keys
         - max damages in euro / m road length
        """

        # Note, these values are in euro/km; while RA2CE standard unit is euro/m length

        lookup_dict = OrderedDict(
            [
                (
                    "motorway",
                    {1: 175000, 2: 350000, 3: 450000, 4: 550000, 5: 650000, 6: 750000},
                ),
                (
                    "trunk",
                    {1: 175000, 2: 300000, 3: 400000, 4: 475000, 5: 575000, 6: 650000},
                ),
                (
                    "primary",
                    {1: 125000, 2: 250000, 3: 325000, 4: 425000, 5: 500000, 6: 575000},
                ),
                (
                    "secondary",
                    {1: 125000, 2: 225000, 3: 300000, 4: 400000, 5: 475000, 6: 550000},
                ),
                (
                    "tertiary",
                    {1: 100000, 2: 175000, 3: 250000, 4: 350000, 5: 425000, 6: 500000},
                ),
                (
                    "track",
                    {1: 75000, 2: 150000, 3: 225000, 4: 300000, 5: 375000, 6: 450000},
                ),
                (
                    "other",
                    {1: 75000, 2: 150000, 3: 225000, 4: 300000, 5: 375000, 6: 450000},
                ),
            ]
        )

        new_dict = {}
        for road_type, lanedicts in lookup_dict.items():
            new_lanedict = {}
            for lane, costs in lanedicts.items():
                new_lanedict[lane] = costs / 1000
            new_dict[road_type] = new_lanedict

        lookup_dict = new_dict

        return lookup_dict

    @staticmethod
    def get_flood_curves() -> dict:
        """Lookup flood curve values and create interpolator around it

        Units of the interpolator objects: water depth in cm on x-axis; damage fraction (unitless) on y-axis
        """
        _depth_cm_str = "depth (cm)"
        _damage_str = "damage (% of total construction costs)"
        lookup_dict = {
            "C1": {
                0: _depth_cm_str,
                1: 0,
                2: 50,
                3: 100,
                4: 150,
                5: 200,
                6: 600,
                7: np.nan,
                8: np.nan,
                9: np.nan,
            },
            "Unnamed: 2": {
                0: _damage_str,
                1: 0,
                2: 0.01,
                3: 0.03,
                4: 0.075,
                5: 0.1,
                6: 0.2,
                7: np.nan,
                8: np.nan,
                9: np.nan,
            },
            "C2": {
                0: _depth_cm_str,
                1: 0,
                2: 50,
                3: 100,
                4: 150,
                5: 200,
                6: 600,
                7: np.nan,
                8: np.nan,
                9: np.nan,
            },
            "Unnamed: 4": {
                0: _damage_str,
                1: 0,
                2: 0.02,
                3: 0.06,
                4: 0.1,
                5: 0.12,
                6: 0.22,
                7: np.nan,
                8: np.nan,
                9: np.nan,
            },
            "C3": {
                0: _depth_cm_str,
                1: 0,
                2: 50,
                3: 100,
                4: 150,
                5: 200,
                6: 600,
                7: np.nan,
                8: np.nan,
                9: np.nan,
            },
            "Unnamed: 6": {
                0: _damage_str,
                1: 0,
                2: 0.002,
                3: 0.004,
                4: 0.025,
                5: 0.03,
                6: 0.04,
                7: np.nan,
                8: np.nan,
                9: np.nan,
            },
            "C4": {
                0: _depth_cm_str,
                1: 0,
                2: 50,
                3: 100,
                4: 150,
                5: 200,
                6: 600,
                7: np.nan,
                8: np.nan,
                9: np.nan,
            },
            "Unnamed: 8": {
                0: _damage_str,
                1: 0,
                2: 0.015,
                3: 0.04,
                4: 0.2,
                5: 0.25,
                6: 0.35,
                7: np.nan,
                8: np.nan,
                9: np.nan,
            },
            "C5": {
                0: _depth_cm_str,
                1: 0,
                2: 50,
                3: 100,
                4: 200,
                5: 600,
                6: np.nan,
                7: np.nan,
                8: np.nan,
                9: np.nan,
            },
            "Unnamed: 10": {
                0: _damage_str,
                1: 0,
                2: 0.015,
                3: 0.025,
                4: 0.035,
                5: 0.05,
                6: np.nan,
                7: np.nan,
                8: np.nan,
                9: np.nan,
            },
            "C6": {
                0: _depth_cm_str,
                1: 0,
                2: 50,
                3: 100,
                4: 200,
                5: 600,
                6: np.nan,
                7: np.nan,
                8: np.nan,
                9: np.nan,
            },
            "Unnamed: 12": {
                0: _damage_str,
                1: 0,
                2: 0.12,
                3: 0.2,
                4: 0.28,
                5: 0.35,
                6: np.nan,
                7: np.nan,
                8: np.nan,
                9: np.nan,
            },
            "HZ": {
                0: _depth_cm_str,
                1: 0.0,
                2: 50.0,
                3: 100.0,
                4: 150.0,
                5: 200.0,
                6: 300.0,
                7: 400.0,
                8: 500.0,
                9: 600.0,
            },
            "Unnamed: 14": {
                0: _damage_str,
                1: 0.0,
                2: 0.25,
                3: 0.42,
                4: 0.55,
                5: 0.65,
                6: 0.8,
                7: 0.9,
                8: 1.0,
                9: 1.0,
            },
        }

        flood_curves = pd.DataFrame.from_dict(lookup_dict)
        headers = flood_curves.columns

        # Convert to ra2ce standard units (depth (m))
        depth_cols = [
            col for col in flood_curves.columns if not col.startswith("Unnamed")
        ]
        flood_curves.loc[1:, depth_cols] = flood_curves.loc[1:, depth_cols] / 100
        flood_curves.loc[0, depth_cols] = "depth (m)"

        curve_name = [0] * int(len(headers) / 2)  # create empty arrays
        interpolators = [0] * int(len(headers) / 2)

        for i in range(0, int(len(headers) / 2)):  # iterate over the damage curves
            curve_name[i] = headers[i * 2]
            curve = flood_curves.iloc[:, 2 * i : 2 * i + 2].dropna()
            # curve x-values in the even; and y-values in the uneven columns
            interpolators[i] = interp1d(
                curve.values[1:, 0],
                curve.values[1:, 1],
                fill_value=(curve.values[1, 1], curve.values[-1, 1]),
                bounds_error=False,
            )
        return OrderedDict(zip(curve_name, interpolators))


class CreateLookupTables:
    """This class lets you create the dictionary lookup tables from an excel file.
    ONLY use this class if dictionairies need to be updated."""

    def __init__(self, settings_file: Path):
        self.settings_file = settings_file

    def create(self):
        lane_damage_correction = self.load_lane_damage_correction("Max_damages", "G:M")
        dict_max_damages = self.import_damage("Max_damages", usecols="C:E")
        max_damages_hz = self.load_hz_max_dam("Huizinga_max_dam", "A:G")
        interpolators = self.import_flood_curves(sheet_name="All_curves", usecols="B:O")

        return (
            lane_damage_correction,
            dict_max_damages,
            max_damages_hz,
            interpolators,
        )

    def load_hz_max_dam(self, sheet_name, usecols):
        """
        Loads the maximum damages according to Huizinga from an Excel file

        Argument:
            *filename* (string) - name of the Excel file (should be located in the input_path dir)
            *sheet_name* (string) - name of the excel sheet name
            *usecols* (string) - the columns which have the data (first column should have the road_type keys)

        Returns:
            *HZ_max_dam* (OrderedDict) - keys are road_types; values are dicts with key: lane, value = correction factor
                Use like: lane_corr['motorway'][4] -> 1.25 (i.e. correct max damage by +25%)
        """

        lane_corr_df = pd.read_excel(
            self.settings_file,
            sheet_name=sheet_name,
            header=0,
            usecols=usecols,
            index_col=0,
        )
        odf = OrderedDict()  # initialize OrderedDict
        hz_max_dam = lane_corr_df.to_dict(orient="index", into=odf)
        return hz_max_dam

    def import_damage(self, sheet_name, usecols):
        """
        Imports the maximum damage data from an Excel file in the input_path folder

        Arguments:
            *file_name* (string) : name of the Excel file (should be located in the input_path folder)
            *sheet_name* (string) : name of the Excel sheet containing the data
            *usecols* (string) : columns containing the data you want to read, including the column with the road_types e.g. "C:F"

        Returns:
            *dict* (Ordered Dictionary) : An ordered dictionary with a group of damage estimates as keys;
                 each value contains another ordered dictionary with as keys the types of roads and as values the damages in Euros
                    So you call the output as: dict['Worldbank'] to get a dict with all the damages in WorldBank
                    And dict['Worldbank']['motorway'] to get the damage for a motorway according to the worldbank

                    #From version 0.7 and higher, this structure maybe does not make much sense, because we use upper and lower bounds

        """

        df = pd.read_excel(
            self.settings_file,
            sheet_name=sheet_name,
            header=[3],
            usecols=usecols,
            index_col=0,
        )
        df = df.iloc[df.index.notna(), :]  # Drop the empty cells
        odf = OrderedDict()  # initialize OrderedDict
        return df.to_dict(into=odf)

    def load_lane_damage_correction(self, sheet_name, usecols):
        """
        Loads the maximum damage correction from an Excel file into an ordered dict.

        Argument:
            *filename* (string) - name of the Excel file (should be located in the input_path dir)
            *sheet_name* (string) - name of the excel sheet name
            *usecols* (string) - the columns which have the data (first column should have the road_type keys)

        Returns:
            *lane_corr* (OrderedDict) - keys are road_types; values are dicts with key: lane, value = correction factor
                Use like: lane_corr['motorway'][4] -> 1.25 (i.e. correct max damage by +25%)
        """

        lane_corr_df = pd.read_excel(
            self.settings_file,
            sheet_name=sheet_name,
            header=3,
            usecols=usecols,
            index_col=0,
        )
        odf = OrderedDict()  # initialize OrderedDict
        lane_corr = lane_corr_df.to_dict(orient="index", into=odf)
        return lane_corr

    def import_flood_curves(self, sheet_name, usecols):
        """
        Imports the flood curves from a predefined path

        Arguments:
            *filename* (string) : name of the Excel file (should be located in the input_path folder) e.g. "Costs_curves_Europe.xlsx"
            *sheet_name* (string) : name of the Excel sheet containing the damage curves (e.g. 'curves')
            *usecols* (string) : string with the columns of the Excel sheet you want to import, e.g. "B:AA"

        Returns:
            *OrderedDict* : keys are the names of the damage curves
                            values are scipy interpolators
        """

        flood_curves_old = pd.read_excel(
            self.settings_file,
            sheet_name=sheet_name,
            header=[2],
            index_col=None,
            usecols=usecols,
        )  # removed skip-footer; gave unexpected results
        flood_new = flood_curves_old.to_dict()
        flood_curves = pd.DataFrame.from_dict(flood_new)

        headers = flood_curves.columns
        curve_name = [0] * int(len(headers) / 2)  # create empty arrays
        interpolators = [0] * int(len(headers) / 2)
        for i in range(
            0, int(len(headers) / 2)
        ):  # iterate over the damage curves in the Excel file
            curve_name[i] = headers[i * 2]
            curve = flood_curves.iloc[:, 2 * i : 2 * i + 2].dropna()
            # curve x-values in the even; and y-values in the uneven columns
            interpolators[i] = interp1d(
                curve.values[1:, 0],
                curve.values[1:, 1],
                fill_value=(curve.values[1, 1], curve.values[-1, 1]),
                bounds_error=False,
            )
        return OrderedDict(zip(curve_name, interpolators))



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007

    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import logging
from copy import deepcopy
from pathlib import Path

from geopandas import GeoDataFrame

from ra2ce.analysis.analysis_result_wrapper import AnalysisResultWrapper


class AnalysisResultWrapperExporter:
    def export_result(
        self,
        result_wrapper: AnalysisResultWrapper,
    ):
        """
        Exports the given result into the analysis requested formats ( `.gpkg` and / or `.csv`).

        Args:
            result_wrapper (AnalysisResultWrapper): The result to export.
        """
        if not result_wrapper.is_valid_result():
            return

        _analysis = result_wrapper.analysis
        _output_path = _analysis.output_path.joinpath(
            _analysis.analysis.analysis.config_value
        )
        _analysis_name = _analysis.analysis.name.replace(" ", "_")

        if _analysis.analysis.save_gpkg:
            self._export_gdf(
                result_wrapper.analysis_result,
                _output_path.joinpath(_analysis_name + ".gpkg"),
            )
        if _analysis.analysis.save_csv:
            self._export_csv(
                result_wrapper.analysis_result,
                _output_path.joinpath(_analysis_name + ".csv"),
            )

    def _export_gdf(self, gdf: GeoDataFrame, export_path: Path):
        """Takes in a geodataframe object and outputs shapefiles at the paths indicated by edge_shp and node_shp

        Arguments:
            gdf [geodataframe]: geodataframe object to be converted
            export_path [Path]: path to save
        """
        # save to shapefile
        gdf.crs = "epsg:4326"  # TODO: decide if this should be variable with e.g. an output_crs configured

        for col in gdf.columns:
            if gdf[col].dtype == object and col != gdf.geometry.name:
                gdf[col] = gdf[col].astype(str)

        if export_path.exists():
            export_path.unlink()
        if not export_path.parent.exists():
            export_path.parent.mkdir(parents=True)

        gdf.to_file(export_path, driver="GPKG")
        logging.info("Results saved to: %s", export_path)

    def _export_csv(self, result_gdf: GeoDataFrame, export_path: Path):
        if not export_path.parent.exists():
            export_path.parent.mkdir(parents=True)

        _result_copy = deepcopy(result_gdf)
        del _result_copy["geometry"]
        _result_copy.to_csv(export_path, index=False)



from ra2ce.analysis.losses.losses_base import LossesBase
from ra2ce.analysis.losses.multi_link_redundancy import MultiLinkRedundancy


class MultiLinkLosses(LossesBase):
    """
    Calculates the multi-link redundancy losses of a NetworkX graph.

    The function removes all links of a variable that have a minimum value
    of min_threshold. For each link it calculates the alternative path, if any available.
    This function only removes one group at the time and saves the data from removing that group.

    This class is based on the LossesBase abstract base class.
    Don't override other methods than _get_criticality_analysis.
    """

    def _get_criticality_analysis(self) -> MultiLinkRedundancy:
        return MultiLinkRedundancy(self.analysis_input)



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007

    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from __future__ import annotations

import logging

from ra2ce.analysis.analysis_config_wrapper import AnalysisConfigWrapper
from ra2ce.network.network_config_wrapper import NetworkConfigWrapper


class ConfigWrapper:
    network_config: NetworkConfigWrapper
    analysis_config: AnalysisConfigWrapper

    def __init__(self) -> None:
        self.network_config = None
        self.analysis_config = None

    def is_valid_input(self) -> bool:
        """
        Validates whether the input is valid. This require that at least the analysis ini file is given.
        TODO: Very unclear what a valid input is, needs to be better specified.

        Returns:
            bool: Input parameters are valid for a ra2ce run.
        """
        if not self.analysis_config or not self.analysis_config.is_valid():
            logging.error("No valid analyses.ini file provided. Program will close.")
            return False

        if self.network_config and not self.network_config.is_valid():
            logging.error("No valid network.ini file provided. Program will close.")
            return False

        if (
            self.network_config
            and self.analysis_config
            and (
                self.analysis_config.config_data.root_path
                != self.network_config.config_data.root_path
            )
        ):
            logging.error(
                "Root directory differs between network and analyses .ini files"
            )
            return False

        return True

    def configure(self) -> None:
        if self.network_config:
            self.network_config.configure()
        if self.analysis_config:
            self.analysis_config.configure()



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007

    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from __future__ import annotations

from pathlib import Path

from ra2ce.analysis.analysis_config_data.analysis_config_data import AnalysisConfigData
from ra2ce.analysis.analysis_config_data.analysis_config_data_validator import (
    AnalysisConfigDataValidator,
)
from ra2ce.common.configuration.config_wrapper_protocol import ConfigWrapperProtocol
from ra2ce.network.graph_files.graph_files_collection import GraphFilesCollection
from ra2ce.network.network_config_wrapper import NetworkConfigWrapper


class AnalysisConfigWrapper(ConfigWrapperProtocol):
    ini_file: Path
    config_data: AnalysisConfigData
    graph_files: GraphFilesCollection

    def __init__(self) -> None:
        self.ini_file = None
        self.config_data = AnalysisConfigData()
        self.graph_files = GraphFilesCollection()

    def initialize_output_dirs(self) -> None:
        """
        #Initializes the required output directories for a Ra2ce analysis.
        """
        # Create the output folders
        for a in self.config_data.analyses:
            output_path = self.config_data.output_path.joinpath(a.analysis.config_value)
            output_path.mkdir(parents=True, exist_ok=True)

    @classmethod
    def from_data_with_network(
        cls,
        ini_file: Path,
        config_data: AnalysisConfigData,
        network_config: NetworkConfigWrapper,
    ) -> AnalysisConfigWrapper:
        """
        Initializes an `AnalysisConfigWrapper` with the given parameters,
        with network_config as input from the network phase.

        Args:
            ini_file (Path): Path to the ini file containing the analysis data.
            config_data (AnalysisIniConfigData): Ini data representation.
            network_config (NetworkConfig): Network configuration to be used on this analysis.

        Raises:
            FileNotFoundError: When the provided `ini file` cannot be found.

        Returns:
            AnalysisConfigWrapper: Initialized instance.
        """
        if not ini_file.is_file():
            raise FileNotFoundError(ini_file)
        _new_analysis = cls()
        _new_analysis.ini_file = ini_file
        _new_analysis.config_data = config_data

        if not isinstance(network_config, NetworkConfigWrapper):
            raise ValueError("No valid network config provided.")
        _new_analysis.config_data.network = network_config.config_data.network
        _new_analysis.config_data.origins_destinations = (
            network_config.config_data.origins_destinations
        )
        # Graphs are retrieved from the already configured object
        _new_analysis.graph_files = network_config.graph_files

        return _new_analysis

    def configure(self) -> None:
        self.initialize_output_dirs()

    def is_valid(self) -> bool:
        return AnalysisConfigDataValidator(self.config_data).validate().is_valid()



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007

    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from __future__ import annotations

import logging
from pathlib import Path
from typing import Optional


class Ra2ceLogger:
    """
    Class to handle the logging configuration for the RA2CE application.
    """

    log_file: Optional[Path] = None
    _file_handler: Optional[logging.FileHandler] = None

    def __init__(self, log_file: Optional[Path], logger_name: str) -> None:
        self.log_file = log_file
        self._set_file_handler()
        self._set_console_handler()
        self._set_formatter()
        logging.info(f"{logger_name} logger initialized.")

    @classmethod
    def initialize_file_logger(cls, logging_dir: Path, logger_name: str) -> Ra2ceLogger:
        """
        Initializes a logger that writes to a file and to console.

        Args:
            logging_dir (Path): Path to the logging directory
            logger_name (str): Name of the logger

        Returns:
            Ra2ceLogger: The logger object
        """
        if not logging_dir.is_dir():
            logging_dir.mkdir(parents=True)
        _log_file = logging_dir.joinpath(f"{logger_name}.log")
        if not _log_file.is_file():
            _log_file.touch()
        return cls(_log_file, logger_name)

    @classmethod
    def initialize_console_logger(cls, logger_name: str) -> Ra2ceLogger:
        """
        Initializes a logger that writes to console only.

        Args:
            logger_name (str): Name of the logger

        Returns:
            Ra2ceLogger: The logger object
        """
        return cls(None, logger_name)

    def _get_logger(self) -> logging.Logger:
        """
        Gets the ra2ce logger which by default is the root logging.Logger.

        Returns:
            logging.Logger: Logger instance.
        """
        return logging.getLogger("")

    def _set_file_handler(self) -> None:
        # Create a root logger and set the minimum logging level.
        if not self.log_file:
            return
        self._get_logger().setLevel(logging.INFO)
        self._file_handler = logging.FileHandler(filename=self.log_file, mode="a")
        self._file_handler.setLevel(logging.INFO)
        self._get_logger().addHandler(self._file_handler)

    def _set_console_handler(self) -> None:
        # Create a console handler and set the required logging level.
        self._console_handler = logging.StreamHandler()
        self._console_handler.setLevel(logging.INFO)  # Can be also set to WARNING
        self._get_logger().addHandler(self._console_handler)

    def _set_formatter(self) -> None:
        # Create a formatter and add to the file and console handlers.
        _formatter = logging.Formatter(
            fmt="%(asctime)s - [%(filename)s:%(lineno)d] - %(name)s - %(levelname)s - %(message)s",
            datefmt="%Y-%m-%d %I:%M:%S %p",
        )
        if self.log_file:
            self._file_handler.setFormatter(_formatter)
        self._console_handler.setFormatter(_formatter)



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007

    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


from pathlib import Path
from typing import Any, Protocol, runtime_checkable


@runtime_checkable
class FileReaderProtocol(Protocol):
    def read(self, file_path: Path) -> Any:
        """
        Reads from a given file and converts the data into `Any` type.

        Args:
            file_path (Path): Path to a file containing data to be read.

        Returns:
            Any: Object mapped from the data in the file.
        """
        pass



from ra2ce.analysis.losses.losses_base import LossesBase
from ra2ce.analysis.losses.single_link_redundancy import SingleLinkRedundancy


class SingleLinkLosses(LossesBase):
    """
    Calculates the single-link redundancy losses of a NetworkX graph.
    This is the function to analyse roads with a single link disruption and an alternative route.

    This class is based on the LossesBase abstract base class.
    Don't override other methods than _get_criticality_analysis.
    """

    def _get_criticality_analysis(self) -> SingleLinkRedundancy:
        return SingleLinkRedundancy(self.analysis_input)



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007

    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import logging
from pathlib import Path
from typing import Any

import geopandas as gpd
import momepy
import networkx as nx
import pandas as pd
import pyproj
from shapely.geometry import Point
from tqdm import tqdm

import ra2ce.network.networks_utils as nut
from ra2ce.network.avg_speed.avg_speed_calculator import AvgSpeedCalculator
from ra2ce.network.exporters.json_exporter import JsonExporter
from ra2ce.network.network_config_data.network_config_data import NetworkConfigData
from ra2ce.network.network_simplification import NetworkGraphSimplificator
from ra2ce.network.network_wrappers.network_wrapper_protocol import (
    NetworkWrapperProtocol,
)


class VectorNetworkWrapper(NetworkWrapperProtocol):
    """A class for handling and manipulating vector files.

    Provides methods for reading vector data, cleaning it, and setting up graph and
    network.
    """

    def __init__(
        self,
        config_data: NetworkConfigData,
    ) -> None:
        self.attributes_to_exclude_in_simplification = (
            config_data.network.attributes_to_exclude_in_simplification
        )
        self.crs = config_data.crs

        # Network options
        self.primary_files = config_data.network.primary_file
        self.directed = config_data.network.directed

        # Origins Destinations
        self.region_path = config_data.origins_destinations.region
        self.file_id = config_data.network.file_id
        self.link_type_column = config_data.network.link_type_column
        self.output_graph_dir = config_data.output_graph_dir

        # Cleanup
        self.delete_duplicate_nodes = config_data.cleanup.delete_duplicate_nodes

    def get_network(
        self,
    ) -> tuple[nx.Graph, gpd.GeoDataFrame]:
        """Gets a network built from vector files.

        Returns:
            nx.MultiGraph: MultiGraph representing the graph.
            gpd.GeoDataFrame: GeoDataFrame representing the network.
        """
        gdf = self._read_vector_to_project_region_and_crs()
        gdf = self.clean_vector(gdf)
        if self.directed:
            graph = self._get_direct_graph_from_vector(
                gdf=gdf, edge_attributes_to_include=["avgspeed", "bridge", "tunnel"]
            )
        else:
            graph = self._get_undirected_graph_from_vector(
                gdf, edge_attributes_to_include=["avgspeed", "bridge", "tunnel"]
            )
        edges, nodes = self.get_network_edges_and_nodes_from_graph(graph)
        graph_complex = nut.graph_from_gdf(edges, nodes, node_id="node_fid")
        if self.delete_duplicate_nodes:
            graph_complex = self._delete_duplicate_nodes(graph_complex)

        logging.info("Start converting the complex graph to a simple graph")
        # Create 'graph_simple'
        graph_simple, graph_complex, link_tables = NetworkGraphSimplificator(
            graph_complex=graph_complex,
            attributes_to_exclude=self.attributes_to_exclude_in_simplification,
        ).simplify()

        # Assign the average speed and time to the graphs
        graph_simple = AvgSpeedCalculator(graph_simple, self.output_graph_dir).assign()
        graph_complex = AvgSpeedCalculator(
            graph_complex, self.output_graph_dir
        ).assign()

        logging.info("Start converting the graph to a geodataframe")
        edges_complex, _ = nut.graph_to_gdf(graph_complex)
        logging.info("Finished converting the graph to a geodataframe")

        # Save the link tables linking complex and simple IDs
        self._export_linking_tables(link_tables)

        if not self.directed and isinstance(graph_simple, nx.MultiDiGraph):
            graph_simple = graph_simple.to_undirected()

        # Check if all geometries between nodes are there, if not, add them as a straight line.
        graph_simple = nut.add_missing_geoms_graph(graph_simple, geom_name="geometry")
        logging.info("Finished converting the complex graph to a simple graph")

        return graph_simple, edges_complex

    def _read_vector_to_project_region_and_crs(self) -> gpd.GeoDataFrame:
        gdf = self._read_files(self.primary_files)
        if gdf is None:
            logging.info("no file is read.")
            return None

        # set crs and reproject if needed
        if not gdf.crs and self.crs:
            gdf = gdf.set_crs(self.crs)
            logging.info("setting crs as default EPSG:4326. specify crs if incorrect")

        if self.crs:
            gdf = gdf.to_crs(self.crs)
            logging.info("reproject vector file to project crs")

        # clip for region
        if self.region_path:
            _region_gpd = self._read_files([self.region_path])
            gdf = gpd.overlay(gdf, _region_gpd, how="intersection", keep_geom_type=True)
            logging.info("clip vector file to project region")

        # validate
        if not any(gdf):
            logging.warning("No vector features found within project region")
            return None

        return gdf

    def _read_files(self, file_list: list[Path]) -> gpd.GeoDataFrame:
        """Reads a list of files into a GeoDataFrame.

        Args:
            file_list (list[Path]): List of file paths.

        Returns:
            gpd.GeoDataFrame: GeoDataFrame representing the data.
        """
        # read file
        gdf = gpd.GeoDataFrame(
            pd.concat([gpd.read_file(_fl, engine="pyogrio") for _fl in file_list])
        )
        logging.info(
            "Read files {} into a 'GeoDataFrame'.".format(
                ", ".join(map(str, file_list))
            )
        )
        return gdf

    @staticmethod
    def _delete_duplicate_nodes(graph_complex: nx.MultiGraph) -> nx.MultiGraph:
        # Create a mapping from location to a representative node
        location_to_representative = {}

        # Iterate through nodes and add representative nodes to the mapping
        for node, data in tqdm(
            graph_complex.nodes(data=True),
            desc="Creating graph with removed node duplicates",
        ):
            location = (round(data["geometry"].x, 7), round(data["geometry"].y, 7))
            if location not in location_to_representative:
                location_to_representative[location] = node

        # Create an updated graph with representative nodes
        updated_graph = nx.MultiGraph()

        updated_graph.add_nodes_from(
            (representative_node, {"geometry": Point(location)})
            for location, representative_node in location_to_representative.items()
        )

        # Add edges to the updated_graph
        for u, v, data in tqdm(
            graph_complex.edges(data=True), desc="Adding edges to the updated graph"
        ):
            u_representative = location_to_representative.get(
                (
                    round(graph_complex.nodes[u]["geometry"].x, 7),
                    round(graph_complex.nodes[u]["geometry"].y, 7),
                )
            )
            v_representative = location_to_representative.get(
                (
                    round(graph_complex.nodes[v]["geometry"].x, 7),
                    round(graph_complex.nodes[v]["geometry"].y, 7),
                )
            )

            if u_representative is not None and v_representative is not None:
                # Add representative nodes if they don't exist in the updated graph
                if not updated_graph.has_node(u_representative):
                    updated_graph.add_node(
                        u_representative,
                        geometry=Point(
                            (
                                round(graph_complex.nodes[u]["geometry"].x, 7),
                                round(graph_complex.nodes[u]["geometry"].y, 7),
                            )
                        ),
                    )

                if not updated_graph.has_node(v_representative):
                    updated_graph.add_node(
                        v_representative,
                        geometry=Point(
                            (
                                round(graph_complex.nodes[v]["geometry"].x, 7),
                                round(graph_complex.nodes[v]["geometry"].y, 7),
                            )
                        ),
                    )

                # Add edges to the updated graph
                updated_graph.add_edge(u_representative, v_representative, **data)

        # Set the CRS for updated_graph equal to the CRS of graph_complex
        updated_graph.graph["crs"] = graph_complex.graph.get(
            "crs", pyproj.CRS("EPSG:4326")
        )
        updated_graph.graph["name"] = graph_complex.graph.get("name", None)
        return updated_graph

    def _create_graph_from_gdf(
        self,
        geo_dataframe: gpd.GeoDataFrame,
        edge_attributes_to_include: list,
    ) -> nx.Graph | nx.DiGraph:
        """
        Creates a simple undirected graph with node and edge geometries based on a given GeoDataFrame.

        Args:
            gdf (gpd.GeoDataFrame): Input GeoDataFrame containing line geometries.
                Allow both LineString and MultiLineString.
            edge_attributes_to_include (List[str], optional): Additional attributes to include from the GeoDataFrame in the graph.

        Returns:
            nx.Graph: NetworkX graph object with node and edge geometries and specified attributes.
        """
        _networkx_graph = nx.DiGraph(crs=geo_dataframe.crs, approach="primal")
        for _, row in geo_dataframe.iterrows():
            link_id = row.get(self.file_id, None)
            link_type = row.get(self.link_type_column, None)

            from_node = row.geometry.coords[0]
            to_node = row.geometry.coords[-1]
            _edge_attributes = {
                f"{self.file_id}": link_id,
                f"{self.link_type_column}": link_type,
                "avgspeed": row.pop("avgspeed") if "avgspeed" in row else None,
                "geometry": row.pop("geometry"),
            }
            _networkx_graph.add_node(from_node, geometry=Point(from_node))
            _networkx_graph.add_node(to_node, geometry=Point(to_node))
            _networkx_graph.add_edge(
                from_node,
                to_node,
                link_id=link_id,
                **_edge_attributes,
            )
            if edge_attributes_to_include:
                for edge_attribute_to_include in edge_attributes_to_include:
                    edge_attribute = (
                        row[edge_attribute_to_include]
                        if edge_attribute_to_include in row
                        else None
                    )
                    if edge_attribute:
                        edge_data = _networkx_graph[from_node][to_node]
                        edge_data[edge_attribute_to_include] = edge_attribute

        return _networkx_graph

    def _get_direct_graph_from_vector(
        self, gdf: gpd.GeoDataFrame, edge_attributes_to_include: list
    ) -> nx.DiGraph:
        """Creates a simple directed graph with node and edge geometries based on a given GeoDataFrame.

        Args:
            gdf (gpd.GeoDataFrame): Input GeoDataFrame containing line geometries.
                Allow both LineString and MultiLineString.
            edge_attributes_to_include: Attributes needed to be included from gdf in the graph


        Returns:
            nx.DiGraph: NetworkX graph object with "crs", "approach" as graph properties.
        """

        # simple geometry handling
        gdf = VectorNetworkWrapper.explode_and_deduplicate_geometries(gdf)

        # to graph
        return self._create_graph_from_gdf(gdf, edge_attributes_to_include)

    def _get_undirected_graph_from_vector(
        self, gdf: gpd.GeoDataFrame, edge_attributes_to_include: list
    ) -> nx.Graph:
        """Creates a simple undirected graph with node and edge geometries based on a given GeoDataFrame.

        Args:
            gdf (gpd.GeoDataFrame): Input GeoDataFrame containing line geometries.
                Allow both LineString and MultiLineString.
            edge_attributes_to_include: Attributes needed to be included from gdf in the graph

        Returns:
            nx.Graph: NetworkX graph object with "crs", "approach" as graph properties.
        """
        digraph = self._get_direct_graph_from_vector(
            gdf=gdf, edge_attributes_to_include=edge_attributes_to_include
        )
        return digraph.to_undirected()

    @staticmethod
    def get_network_edges_and_nodes_from_graph(
        graph: nx.Graph,
    ) -> tuple[gpd.GeoDataFrame, gpd.GeoDataFrame]:
        """Sets up network nodes and edges from a given graph.

        Args:
            graph (nx.Graph): Input graph with geometry for nodes and edges.
                Must contain "crs" as graph property.

        Returns:
            gpd.GeoDataFrame: GeoDataFrame representing the network edges with "edge_fid", "node_A", and "node_B".
            gpd.GeoDataFrame: GeoDataFrame representing the network nodes with "node_fid".
        """

        # TODO ths function use conventions. Good to make consistant convention with osm
        nodes, edges = momepy.nx_to_gdf(graph, nodeID="node_fid")
        edges["edge_fid"] = (
            edges["node_start"].astype(str) + "_" + edges["node_end"].astype(str)
        )
        edges.rename(
            {"node_start": "node_A", "node_end": "node_B"}, axis=1, inplace=True
        )
        if not nodes.crs:
            nodes.crs = graph.graph["crs"]
        if not edges.crs:
            edges.crs = graph.graph["crs"]

        if "fid" in edges.columns:
            edges = edges.drop(columns=["fid"])
        return edges, nodes

    @staticmethod
    def clean_vector(gdf: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
        """Cleans a GeoDataFrame.

        Args:
            gdf (gpd.GeoDataFrame): Input GeoDataFrame.

        Returns:
            gpd.GeoDataFrame: Cleaned GeoDataFrame.
        """

        gdf = VectorNetworkWrapper.explode_and_deduplicate_geometries(gdf)

        return gdf

    @staticmethod
    def explode_and_deduplicate_geometries(gpd: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
        """Explodes and deduplicates geometries a GeoDataFrame.

        Args:
            gpd (gpd.GeoDataFrame): Input GeoDataFrame.

        Returns:
            gpd.GeoDataFrame: GeoDataFrame with exploded and deduplicated geometries.
        """
        gpd = gpd.explode()
        gpd = gpd[
            gpd.index.isin(
                gpd.geometry.apply(lambda geom: geom.wkb).drop_duplicates().index
            )
        ]
        return gpd

    def _export_linking_tables(self, linking_tables: tuple[Any]) -> None:
        _exporter = JsonExporter()
        _exporter.export(
            self.output_graph_dir.joinpath("simple_to_complex.json"), linking_tables[0]
        )
        _exporter.export(
            self.output_graph_dir.joinpath("complex_to_simple.json"), linking_tables[1]
        )
